/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Multifox.
 *
 * The Initial Developer of the Original Code is
 * Jeferson Hultmann <hultmann@gmail.com>
 * Portions created by the Initial Developer are Copyright (C) 2009
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

"use strict";

const EXPORTED_SYMBOLS = ["BrowserWindow", // new-window.js
                         ];
Components.utils.import("resource://multifox-modules/new-window.js");
Components.utils.import("resource://gre/modules/XPCOMUtils.jsm");
Components.utils.import("resource://gre/modules/Services.jsm");

console.log("===>LOADING main.js");


// tabbrowser attributes:
// "multifox-tabbrowser-last-tab"

// tab attributes:
// "multifox-tab-id-provider-tld-enc"
// "multifox-login-submit-user"
// "multifox-tab-error" - sandbox error, page not supported

/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Multifox.
 *
 * The Initial Developer of the Original Code is
 * Jeferson Hultmann <hultmann@gmail.com>
 * Portions created by the Initial Developer are Copyright (C) 2009
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

var m_runner = null;

const BrowserWindow = {

  register: function(win) {
    console.log("BrowserWindow.register");

    if (m_runner === null) {
      // first multifox tab!
      m_runner = new MultifoxRunner();
    }


    // some MultifoxContentEvent_* listeners are not called when
    // there are "unload" listeners with useCapture=true. o_O
    // But they are called if event listener is an anonymous function.
    win.addEventListener("unload", ChromeWindowEvents, false);
    win.addEventListener("activate", ChromeWindowEvents, false);
    win.addEventListener(m_runner.eventSentByContent, onContentEvent, false, true);
    win.addEventListener("mousedown", RedirDetector.onMouseDown, false); // command/click listeners can be called after network request
    win.addEventListener("keydown", RedirDetector.onKeyDown, false);

    var tabbrowser = win.getBrowser();
    tabbrowser.addEventListener("pageshow", updateNonNetworkDocuments, false);

    var container = tabbrowser.tabContainer;
    container.addEventListener("TabSelect", TabContainerEvents, false);
    container.addEventListener("TabClose", TabContainerEvents, false);
    container.addEventListener("SSTabRestoring", TabContainerEvents, false);

    // restore icon after toolbar customization
    var toolbox = win.document.getElementById("navigator-toolbox");
    toolbox.addEventListener("DOMNodeInserted", customizeToolbar, false);


    win.document.documentElement.setAttribute("multifox-window-uninitialized", "true");
  },


  unregister: function(win) {
    console.log("BrowserWindow.unregister");

    win.removeEventListener("unload", ChromeWindowEvents, false);
    win.removeEventListener("activate", ChromeWindowEvents, false);
    win.removeEventListener(m_runner.eventSentByContent, onContentEvent, false);
    win.removeEventListener("mousedown", RedirDetector.onMouseDown, false);
    win.removeEventListener("keydown", RedirDetector.onKeyDown, false);

    var tabbrowser = win.getBrowser();
    tabbrowser.removeEventListener("pageshow", updateNonNetworkDocuments, false);

    var container = tabbrowser.tabContainer;
    container.removeEventListener("TabSelect", TabContainerEvents, false);
    container.removeEventListener("TabClose", TabContainerEvents, false);
    container.removeEventListener("SSTabRestoring", TabContainerEvents, false);

    var toolbox = win.document.getElementById("navigator-toolbox");
    toolbox.removeEventListener("DOMNodeInserted", customizeToolbar, false);

    // TODO last window?
    /*
    if () {
      m_runner.shutdown();
      m_runner = null;
    }
    */
  }
};


var ChromeWindowEvents = {
  handleEvent: function(evt) {
    try {
      this[evt.type](evt);
    } catch (ex) {
      Cu.reportError(ex);
      throw new Error("ChromeWindowEvents exception " + ex);
    }
  },

  unload: function(evt) {
    var win = evt.currentTarget;
    BrowserWindow.unregister(win);
  },

  activate: function(evt) {
    var win = evt.currentTarget;
    var tab = win.getBrowser().selectedTab;
    LoginData.setTabAsDefaultLogin(tab);//BUG init m_welcomeMode, mas falta attr current-tld

    //console.log('win activate', m_welcomeMode);
    LoginData._ensureValid(); // BUG workaround to display welcome icon
    if (m_welcomeMode) {
      updateUI(tab, true);
    }
  }
}


const TabContainerEvents = {
  handleEvent: function(evt) {
    try {
      this[evt.type](evt);
    } catch (ex) {
      Cu.reportError(ex);
      throw new Error("TabContainerEvents exception " + ex);
    }
  },

  TabClose: function(evt) {
    var tab = evt.originalTarget;
    MoveTabWindows.tabCloseSaveId(tab);
  },

  TabSelect: function(evt) {
    var tab = evt.originalTarget;
    MoveTabWindows.tabSelectDetectMove(tab); // moved tab: set id
    updateUI(tab, true);
  },

  SSTabRestoring: function(evt) {
    var tab = evt.originalTarget;
    updateUI(tab, true);
  }
};


// first run?
function onStart() {
  var prefName = "extensions.multifox@hultmann.currentVersion";
  var prefs = Cc["@mozilla.org/preferences-service;1"].getService(Ci.nsIPrefBranch2);
  var ver = prefs.prefHasUserValue(prefName) ? prefs.getCharPref(prefName) : "";
  if (ver === "2.0b3") {
    return;
  }
  prefs.setCharPref(prefName, "2.0b3");


  // TODO ss.deleteWindowValue(doc.defaultView, "multifox-dom-identity-id");

  // remove Multifox 1.x cookies
  var profileId = 2;
  var all;
  do {
    all = removeTldData_cookies("multifox-profile-" + profileId);
    console.log("Migrating: removing cookies 1.x", profileId, ":", all.length);
    profileId++;
  } while ((profileId < 20) || (all.length > 0));


  // remove Multifox 2.x beta 1 cookies
  all = removeTldData_cookies("x-content.x-namespace");
  console.log("Migrating: removing cookies 2.0b1", all.length);

  // remove Multifox 2.x beta 2 cookies
  all = removeTldData_cookies("-.x-namespace");
  console.log("Migrating: removing cookies 2.0b2", all.length);

  // remove Multifox 2.x beta 3 cookies
  //all = removeTldData_cookies("multifox-auth-1");
  //var all2 = removeTldData_cookies("multifox-anon-1");
  //console.log("Migrating: removing cookies 2.0b3", all.length + all2.length);
}


function customizeToolbar(evt) {
  var node = evt.target;
  if ((node.id === "urlbar-container") && (node.parentNode.tagName === "toolbar")) {
    var tab = node.ownerDocument.defaultView.getBrowser().selectedTab;
    updateUI(tab, true);
  }
}


function onContentEvent(evt) {
  evt.stopPropagation();

  var obj = JSON.parse(evt.data);
  var contentDoc = evt.target;
  var rv;
  //console.log('onContentEvent', obj.from);

  switch (obj.from) {
    case "cookie":
      rv = documentCookie(obj, contentDoc);
      //console.log('documentCookie ',obj.cmd,typeof rv,rv,contentDoc.location);
      break;
    case "localStorage":
      rv = windowLocalStorage(obj, contentDoc);
      break;
    case "error":
      showError(contentDoc.defaultView, obj.cmd, "-");
      break;
    default:
      throw new Error("onContentEvent: " + obj.from);
  }

  if (rv === undefined) {
    // no response
    return;
  }

  // send data to content
  var evt2 = contentDoc.createEvent("MessageEvent");
  evt2.initMessageEvent(m_runner.eventSentByChrome, false, false, rv, null, null, null);
  var success = contentDoc.dispatchEvent(evt2);
}


function MultifoxRunner() {
  console.log("MultifoxRunner");
  //m_cookieMgr = Cc["@mozilla.org/cookiemanager;1"].getService(Ci.nsICookieManager2); // TODO


  this._sentByChrome  = "multifox-chrome_event-"  + Math.random().toString(36).substr(2);
  this._sentByContent = "multifox-content_event-" + Math.random().toString(36).substr(2);
  this._inject = new DocStartScriptInjection();

  var obs = Cc["@mozilla.org/observer-service;1"].getService(Ci.nsIObserverService);
  obs.addObserver(SubmitObserver, "earlyformsubmit", false);
  obs.addObserver(LoginData.onCookieRejected, "cookie-rejected", false);
  obs.addObserver(LoginData.onCookieChanged, "cookie-changed", false);

  Cookies.start();
  util.networkListeners.enable(httpListeners.request, httpListeners.response);

  onStart();
}


MultifoxRunner.prototype = {
  get eventSentByChrome() {
    return this._sentByChrome;
  },

  get eventSentByContent() {
    return this._sentByContent;
  },

  shutdown: function() {
    console.log("MultifoxRunner.shutdown");
    util.networkListeners.disable();
    this._inject.stop();
    Cookies.stop();
    LoginData.shutdown();

    var obs = Cc["@mozilla.org/observer-service;1"].getService(Ci.nsIObserverService);
    obs.removeObserver(SubmitObserver, "earlyformsubmit");
    obs.removeObserver(LoginData.onCookieRejected, "cookie-rejected");
    obs.removeObserver(LoginData.onCookieChanged, "cookie-changed");
  }
};


function showError(contentWin, notSupportedFeature, details) {
  var msg = [];
  msg.push("ERROR=" + notSupportedFeature);
  msg.push(details);
  if (contentWin.document) {
    msg.push("location=" + contentWin.document.location);
    if (contentWin.document.documentURIObject) {
      msg.push("uri=     " + contentWin.document.documentURIObject.spec);
    }
  }
  msg.push("title=[" + contentWin.document.title + "]");
  console.log(msg.join("\n"));

  var tab = WindowParents.getTabElement(contentWin);
  tab.setAttribute("multifox-tab-error", notSupportedFeature);
  updateUI(tab, true);
}
/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Multifox.
 *
 * The Initial Developer of the Original Code is
 * Jeferson Hultmann <hultmann@gmail.com>
 * Portions created by the Initial Developer are Copyright (C) 2009
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

/*
attr used:
multifox-invalidate-icon
multifox-tab-id-provider-tld-enc
multifox-tab-id-provider-user-enc
multifox-tab-error
multifox-logging-in
multifox-tab-current-tld

*/

//var MultifoxBox IconBadge = {

var m_welcomeMode = false;

function setWelcomeMode(enable) {
  console.log('setWelcomeMode', enable);
  if (m_welcomeMode === enable) {
    return;
  }
  
  m_welcomeMode = enable; // TODO verificar se alguma aba com attr de login

  var winEnum = Cc["@mozilla.org/appshell/window-mediator;1"].getService(Ci.nsIWindowMediator).getEnumerator("navigator:browser");
  while (winEnum.hasMoreElements()) {
    var win = winEnum.getNext();
    var tab = win.getBrowser().selectedTab;
    updateUI(tab, true);
    if (enable) {
      getIconContainer(win.document).setAttribute("current-error", "welcome"); // force updateStatIcon
    }
  }
}


function updateUI(tab, force) { // force = optional param
try{
  if (tab.hasAttribute("selected") === false) {
    return;
  }

  if (tab.hasAttribute("multifox-invalidate-icon")) {
    tab.removeAttribute("multifox-invalidate-icon");
  } else {
    if (force !== true) {
      return;
    }
  }

  var doc = tab.ownerDocument;
  var container = getIconContainer(doc);

  if ((tab.hasAttribute("multifox-tab-id-provider-tld-enc") === false) && (m_welcomeMode === false)) {
    // remove box
    if (container !== null) {
      container.parentNode.removeChild(container);
    }
    return;
  }

  if (container === null) {
    createContainer(doc);
    var win = doc.defaultView;
    var delay = getTabs(win.getBrowser()).length > 1 ? 25 : 200;
    win.setTimeout(initIcon, delay, doc); // open window => greater delay
  } else {
    LoginData.setTabAsDefaultLogin(tab); // TODO it will call _ensureValid, make sure we don't block TabSelect
    if (container.firstChild !== null) { // waiting initIcon?
      updateIconCore(tab, container);
    }
  }

} catch(ex) {
console.error(ex);
}
}



// <hbox align="center" id="multifox-icon">
//   <box>
//     <hbox align="center">
//       <hbox id="multifox-icon-stat-icon">
//         <image src="warning.png"/>
//       </hbox>
//       <label value="username"/>
//     </hbox>
//   </box>
// </hbox>


function createContainer(doc) {
  var container = getIconContainer(doc);
  if (container === null) {
    var ref = doc.getElementById("urlbar-icons");
    container = ref.insertBefore(doc.createElement("hbox"), ref.firstChild);
    container.setAttribute("hidden", "true");
    container.setAttribute("id", "multifox-icon");
    container.setAttribute("align", "center");
  }
  return container;
}


function createBoxDom(container) {
  var doc = container.ownerDocument;
  var container2 = container.appendChild(doc.createElement("box"));
  var container3 = container2.appendChild(doc.createElement("hbox"));
  var stat = container3.appendChild(doc.createElement("hbox"));
  var label = container3.appendChild(doc.createElement("label"));

  container2.setAttribute("tooltiptext", "Multifox (BETA)");
  container3.setAttribute("align", "center");
  stat.setAttribute("hidden", "true");
  stat.setAttribute("id", "multifox-icon-stat-icon");

  // max-width
  container3.style.maxWidth = "9em"; // TODO util.getText("icon.user.maxWidth");
  label.setAttribute("crop", "end");
  label.setAttribute("flex", "1");
}


function setStyleCore(container, doc) {
  container.minHeight = "16px";
  container.overflow = "hidden";
  container.padding = "0";
  container.margin = "0 2px 0 1px";
  container.border = "1px solid #ccc";
  container.borderRadius = "2px";
  container.MozBorderRadius = "2px"; // Gecko 1.9.2
  container.opacity = "1";
  container.MozTransition = "";

  var styleLabel = getIconLabel(doc).style;
  styleLabel.padding = "0px 6px";
  styleLabel.margin = "-1px 0px";
  styleLabel.color = "#777";
}


function setStyle(mode, containerStyle, iconStyle, labelStyle) {
  switch (mode) {
    case "default":
      containerStyle.backgroundImage = "-moz-linear-gradient(hsl(214,44%,99%), hsl(214,44%,87%))";
      labelStyle.textShadow = "0 1px 0 white";
      iconStyle.opacity = "1";
      break;
    case "hover":
      containerStyle.backgroundImage = "-moz-linear-gradient(hsl(214,44%,80%), hsl(214,44%,60%))";//#fff, #eee)";
      iconStyle.opacity = "1";
      labelStyle.textShadow = "0 1px 0 white, 0 1px 5px white, 3px 0 5px white, 0 -1px 5px white, -3px 0 5px white";
      break;
    case "active":
      containerStyle.backgroundImage = "-moz-linear-gradient(#ddd, #999)";
      iconStyle.opacity = ".5";
      labelStyle.textShadow = "0 1px 0 white, 0 1px 5px white, 3px 0 5px white, 0 -1px 5px white, -3px 0 5px white";
      break;
    default:
      throw new Error("setColor");
  }
}


function invalidateUI(tab) {
  if (tab.hasAttribute("selected")) {
    tab.setAttribute("multifox-invalidate-icon", "true");
  }
}


function updateIconCore(tab, container) {
  if (m_welcomeMode) {
    var doc = tab.ownerDocument;
    getIconLabel(doc).setAttribute("value", "Multifox (BETA)");
    updateStatIcon3(true, getStatIconContainer(doc), "chrome://multifox/content/favicon.ico");
  } else {
    updateStatIcon(tab, container);
    updateIconUserName(tab);
  }
}


function updateIconUserName(tab) {
  var tabLogin = new TabInfo(tab);
  var user;
  if (tabLogin.isNewUser) {
    user = util.getText("icon.add-account.label");
  } else {
    user = tabLogin.plainUser; // BUG sometimes is null
  }

  //console.log("updateIconUserName", user, tabLogin.loginUser16);
  getIconLabel(tab.ownerDocument).setAttribute("value", user);
}


function updateStatIcon(tab, container) {
  var currentError = container.hasAttribute("current-error") ? container.getAttribute("current-error") : "";
  var newError = tab.hasAttribute("multifox-tab-error") ? tab.getAttribute("multifox-tab-error") : "";
  if (currentError !== newError) {
    container.setAttribute("current-error", newError);
    var statIcon = getStatIconContainer(tab.ownerDocument);
    updateStatIcon3(newError.length > 0, statIcon, "chrome://global/skin/icons/warning-16.png"); // ubuntu: 22x22
  }
}


function updateStatIcon3(show, statIcon, url) {
  if (statIcon.firstChild) {
    statIcon.removeChild(statIcon.firstChild);
  }
  if (show) {
    var doc = statIcon.ownerDocument;
    var img = statIcon.appendChild(doc.createElement("image"));
    img.setAttribute("src", url);
    img.setAttribute("width", "16");
    img.setAttribute("height", "16");
    img.style.margin = "0 -4px 0 6px"; // marginleft = same as getIconLabel(doc).style.padding
    statIcon.removeAttribute("hidden");
  } else {
    statIcon.setAttribute("hidden", "true");
  }
}


function getIconContainer(doc) {
  return doc.getElementById("multifox-icon");
}


function getIconLabel(doc) {
  var node = getIconContainer(doc); // TODO receber container -- e nao doc
  console.assert(node !== null, "getIconContainer=null");
  //console.assert(node.firstChild.firstChild !== null, "node.firstChild.firstChild=null");
  return node.firstChild.firstChild.children[1];
}


function getStatIconContainer(doc) {
  return doc.getElementById("multifox-icon-stat-icon");
}


function initIcon(doc) {
  var tab = doc.defaultView.getBrowser().selectedTab;
  LoginData.setTabAsDefaultLogin(tab);

  var container = getIconContainer(doc);
  if (container === null) {
    // e.g. user changed tab!
    console.log("initIcon ignored! " + container);
    return;
  }

  console.assert(container.children.length === 0, "container has children");
  createBoxDom(container);
  setStyleCore(container.style, doc);

  if (tab.hasAttribute("multifox-logging-in")) {
    tab.removeAttribute("multifox-logging-in");
    container.style.opacity = "0";
    container.style.MozTransition = "opacity .3s ease";
    doc.defaultView.setTimeout(function() {
      container.style.opacity = "1";
    }, 50);
  }

  initIconNormal(doc);
  updateIconCore(tab, container);
  container.removeAttribute("hidden");
}



function initIconNormal(doc) {
  var statIcon = getStatIconContainer(doc);
  var container = getIconContainer(doc);
  var lab = getIconLabel(doc).style;

  doc.defaultView.setTimeout(function() {
    // workaround
    // otherwise, mousedown on icon will close and open popup, even with stopPropagation
    statIcon.addEventListener("mousedown", showMsgPanel, false);
    container.addEventListener("mousedown", showMenuPopup, false);
  }, 0);
  container.addEventListener("mouseover", onIconHover, false);
  container.addEventListener("mouseout", onIconHover, false);

  //statIcon.style.opacity = "1";
  setStyle("default", container.style, statIcon.style, lab);
  //lab.opacity = "1";
  //defaultStyle(container.style, lab);
}


function initIconPressed(doc, popupId) {
  var container = getIconContainer(doc);
  container.removeEventListener("mouseover", onIconHover, false);
  container.removeEventListener("mouseout", onIconHover, false);

  var statIcon = getStatIconContainer(doc);
  var lab = getIconLabel(doc).style;
  statIcon.removeEventListener("mousedown", showMsgPanel, false);
  switch (popupId) {
    case "menu":
      setStyle("active", container.style, statIcon.style, lab);
      break;
    case "msg":
      lab.opacity = ".5";
      break;
  }
}


function onIconHover(evt) {
  var doc = evt.target.ownerDocument;
  var container = getIconContainer(doc);
  setStyle(
    evt.type === "mouseover" ? "hover" : "default",
    container.style,
    getStatIconContainer(doc).style,
    getIconLabel(doc).style);
}


function isMsgPopup(evt) {
  return evt.target.localName === "image";
}


function showMsgPanel(evt) {
  if (isMsgPopup(evt) === false) {
    return;
  }

  var doc = evt.target.ownerDocument;
  initIconPressed(doc, "msg");

  var ns = loadSubScript("resource://multifox-modules/popup.js");
  var panel = ns.createMsgPanel(doc);

  panel.addEventListener("popuphidden", function(evt) {
    initIconNormal(doc);
    // remove error msg
    var tab = doc.defaultView.getBrowser().selectedTab;
    tab.removeAttribute("multifox-tab-error");
    updateUI(tab, true);
  }, false);

  panel.openPopup(getStatIconContainer(doc), "after_start", 0, 1);
}


function showMenuPopup(evt) {
  var doc = evt.target.ownerDocument;
  var container = getIconContainer(doc);

  if (m_welcomeMode) {
    var ns = loadSubScript("resource://multifox-modules/welcome.js");
    var panel = ns.welcomePopup(doc);
    panel.openPopup(container, "bottomcenter topright");
    return;
  }

  if (isMsgPopup(evt)) {
    return;
  }

  initIconPressed(doc, "menu");
  var menu = doc.getElementById("mainPopupSet").appendChild(doc.createElement("menupopup"));
  var ns = loadSubScript("resource://multifox-modules/popup.js");
  ns.createLoginsMenu(menu, function() {initIconNormal(doc);});
  menu.openPopup(container, "after_end", 0, 1);
}


function loadSubScript(path) {
  var ns = {};
  var sub = Cc["@mozilla.org/moz/jssubscript-loader;1"].getService(Ci.mozIJSSubScriptLoader);
  sub.loadSubScript(path, ns);
  return ns;
}
/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Multifox.
 *
 * The Initial Developer of the Original Code is
 * Jeferson Hultmann <hultmann@gmail.com>
 * Portions created by the Initial Developer are Copyright (C) 2010
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */


// tab moved (between windows): TabClose + TabSelect
// workaround: check call stack to detect d&d
const StackAnalyzer = {
  swapBrowsersOnTabSelect: function() {
    for (var s = Components.stack; s; s = s.caller) {
      switch (s.name) {
        case "swapBrowsersAndCloseOther":
        case "onxbldrop": //fx4
        case "_onDrop":   //fx36
          return true;
      }
    }
    return false;
  }
};


// keep multifox id when tab is moved between windows
const MoveTabWindows = {
  _lastClosedTab_domain: null,
  _lastClosedTab_user:   null,
  _remainingTabSelect:   0,

  tabCloseSaveId: function(tab) {
    if (tab.hasAttribute("multifox-tab-id-provider-tld-enc")) {
      var tabLogin = new TabInfo(tab);
      this._lastClosedTab_domain = tabLogin.encodedTld;
      this._lastClosedTab_user   = tabLogin.encodedUser;
      this._remainingTabSelect = 2;
    } else {
      this._remainingTabSelect = 0;
    }
  },

  tabSelectDetectMove: function(tab) {
    if (this._remainingTabSelect === 0) {
      return;
    }
    this._remainingTabSelect--;
    if (StackAnalyzer.swapBrowsersOnTabSelect()) {
      this._remainingTabSelect = 0;
      var tabLogin = TabLogin.create(tab, this._lastClosedTab_user, this._lastClosedTab_domain);
      tabLogin.saveToTab();
      console.log('tabSelectDetectMove');
    }
  }

};
/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Multifox.
 *
 * The Initial Developer of the Original Code is
 * Jeferson Hultmann <hultmann@gmail.com>
 * Portions created by the Initial Developer are Copyright (C) 2009
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

// Add hooks to documents (cookie, localStorage, ...)

function DocStartScriptInjection() {
  var is192 = this._is192();
  this._loader = new ScriptSourceLoader(is192);
  var obs = Cc["@mozilla.org/observer-service;1"].getService(Ci.nsIObserverService);
  obs.addObserver(this, this._getTopic(is192), false);
}

DocStartScriptInjection.prototype = {
  stop: function() {
    var obs = Cc["@mozilla.org/observer-service;1"].getService(Ci.nsIObserverService);
    obs.removeObserver(this, this._getTopic(this._is192()));
    delete this._loader;
  },

  _is192: function() {
    var info = Cc["@mozilla.org/xre/app-info;1"].getService(Ci.nsIXULAppInfo);
    return info.platformVersion.indexOf("1.9") === 0; // Gecko 1.9.2
  },

  _getTopic: function(is192) {
    return is192 ? "content-document-global-created" : "document-element-inserted";
  },

  QueryInterface: XPCOMUtils.generateQI([Ci.nsIObserver]),
  observe: function(subject, topic, data) {
    var win;
    var requestUri;

    switch (topic) {
      case "content-document-global-created": // Gecko 1.9.2
        win = subject;
        if (win === null) {
          return;
        }
        if (data === "null") {
          return;
        }
        var io = Cc["@mozilla.org/network/io-service;1"].getService(Ci.nsIIOService);
        requestUri = io.newURI(data, null, null);
        break;
      case "document-element-inserted":
        win = subject.defaultView;
        if (win === null) {
          return; // xsl/xbl
        }
        requestUri = subject.documentURIObject;
        break;
      default:
        throw new Error(topic);
    }

    if (win.wrappedJSObject.hasOwnProperty("_multifox_obs")) {
      console.warn("DUP " + topic + " @ " + requestUri.spec); // https://bugzilla.mozilla.org/show_bug.cgi?id=642145 ?
      return;
    }
    win.wrappedJSObject._multifox_obs = true; // BUG visible by content

    // requestUri=about:neterror?e=dnsNotFound&u=http...
    var tabUser = TabLogin.getFromDomWindow(win);
    if (tabUser === null) {
      return;
    }

    if (tabUser.isLoginInProgress) {
      tabUser = TabLogin.getLoginInProgress(tabUser.tabElement);
    }

    if (isTopWindow(win) === false) {
      // iframe
      //console.log("DocStartScriptInjection iframe", requestUri.spec);
      if (isSupportedScheme(requestUri.scheme)) {
        var tabLogin = getSubElementLogin(requestUri, tabUser, null);
        if ((tabLogin !== null) && tabLogin.isLoggedIn) {
          this._inject(win);
        //} else {
        //  console.log("DocStartScriptInjection NOP", tabLogin, tabLogin !== null, tabLogin?("anonResource" in tabLogin):"?");
        }
      }
      return;
    }

    // new top window
    var tab = tabUser.tabElement;
    if (isSupportedScheme(requestUri.scheme)) {
      if (tabUser.isLoggedIn) {
        this._inject(win);
      }
    } else { // about, javascript
      tab.removeAttribute("multifox-tab-current-tld");
      tabUser.setTabAsAnon();
    }

    updateUI(tab);
    RedirDetector.resetTab(tab);
  },


  _inject: function(win) {
    var sandbox = Cu.Sandbox(win);
    sandbox.window = win.wrappedJSObject;
    sandbox.document = win.document.wrappedJSObject;

    var src = this._loader.getScript();
    try {
      Cu.evalInSandbox(src, sandbox);
    } catch (ex) {
      showError(win, "sandbox", win.document.location + " " + "//exception=" + ex);
    }
  }
};


function isSupportedScheme(scheme) {
  return (scheme === "http") || (scheme === "https");
}


function throwException() {
  console.trace();
  throw new Error("throwException");
}


function ScriptSourceLoader(is192) {
  if (is192) {
    // Gecko 1.9.2
    this._path = "chrome://multifox/content/content-injection-192.js";
  } else {
    // Gecko 2.0
    this._path = "chrome://multifox/content/content-injection.js";
  }
  this._src = null;
  this._load(true);
}


ScriptSourceLoader.prototype = {

  getScript: function() {
    if (this._src === null) {
      this._load(false);
    }
    return this._src;
  },

  _load: function(async) {
    var me = this;
    var xhr = Cc["@mozilla.org/xmlextras/xmlhttprequest;1"].createInstance(Ci.nsIXMLHttpRequest);
    xhr.onload = function() {
      delete me._path;
      me._src = xhr.responseText + "initContext(window, document, '"
                                 + m_runner.eventSentByChrome + "','" + m_runner.eventSentByContent + "');";
    };
    xhr.open("GET", this._path, async);
    xhr.overrideMimeType("text/plain");
    xhr.send(null);
  }
};
/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Multifox.
 *
 * The Initial Developer of the Original Code is
 * Jeferson Hultmann <hultmann@gmail.com>
 * Portions created by the Initial Developer are Copyright (C) 2009
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

const httpListeners = {
  request: {
    QueryInterface: XPCOMUtils.generateQI([Ci.nsIObserver]),

    // nsIObserver
    observe: function REQUEST(aSubject, aTopic, aData) {

try{

      var httpChannel = aSubject.QueryInterface(Ci.nsIHttpChannel);
      var win = getChannelWindow(httpChannel);

      if (win === null) {
        // safebrowsing, http://wpad/wpad.dat
        return;
      }

      var isWin = isWindowChannel(httpChannel);
      var isXhr = false;
      if (httpChannel.notificationCallbacks) {
        isXhr = httpChannel.notificationCallbacks instanceof Ci.nsIXMLHttpRequest;
        try{
          isXhr = httpChannel.notificationCallbacks.getInterface(Ci.nsIXMLHttpRequest) != null;
        }catch(ex){
        }
      }

      /*
      if (isWin || isXhr) {
        console.log("===>request", isWin ? "DOC" : "ELEM", isXhr ? "(XHR)" : "",
                    "\n", httpChannel.requestMethod, httpChannel.URI.spec,
                    "\n", "WIN", win === win.top ? "top" : "frame", win.document.documentURI, win.document.href);
      }
      */

      var tabLogin = getRequestChannelUser(httpChannel, win); // TODO requestLogin
      if ((tabLogin === null) || (tabLogin.isLoggedIn === false)) {
        /*
        console.log("===>request DEFAULT", isWin ? "DOC" : "ELEM", isXhr ? "(XHR)" : "",
                    "\n", httpChannel.requestMethod, httpChannel.URI.spec,
                    "\n", "WIN", win === win.top ? "top" : "frame", win.document.documentURI, win.document.href);
        */
        return; // use default session
      }

      var myHeaders = HttpHeaders.fromRequest(httpChannel);
      if (myHeaders["authorization"] !== null) {
        showError(win, "authorization", "-");
        return;
      }


      RequestLogin.save(tabLogin, httpChannel);


      var cook = Cookies.getCookie(false, httpChannel.URI, tabLogin);
      httpChannel.setRequestHeader("Cookie", cook, false);



}catch(ex){
  console.trace();
  Cu.reportError(ex);
}


    }
  },

  response: {
    QueryInterface: XPCOMUtils.generateQI([Ci.nsIObserver]),

    // nsIObserver
    observe: function RESPONSE(aSubject, aTopic, aData) {
      var httpChannel = aSubject.QueryInterface(Ci.nsIHttpChannel);
      var tabLogin = RequestLogin.get(httpChannel);

      if (tabLogin === null) {
        var win2 = getChannelWindow(httpChannel);
        var myHeaders = HttpHeaders.fromResponse(httpChannel);
        //console.log('response ANON ', (win2?win2.document.location:'NULL?'), myHeaders["set-cookie"], httpChannel.URI.spec);
        RequestLogin.findSavedLogin_byurl(httpChannel.URI.spec);
        return; // use original Set-Cookie
      }

      var myHeaders = HttpHeaders.fromResponse(httpChannel);
      if (myHeaders["www-authenticate"] !== null) {
        var win = getChannelWindow(httpChannel);
        if (win === null) {
          return;
        }
        showError(win, "www-authenticate", "-");
        return;
      }

      var setCookies = myHeaders["set-cookie"];
      if (setCookies === null) {
        return;
      }

      // replace "Set-Cookie"
      httpChannel.setResponseHeader("Set-Cookie", null, false);
      Cookies.setCookie(tabLogin, httpChannel.URI, setCookies, false);
    }
  }
};


const RequestLogin = {
  _logins:   [],
  _channels: [],

  save: function(tabLogin, request) {
    this._logins.push(tabLogin);
    this._channels.push(Cu.getWeakReference(request));
  },

  get: function(response) {
    //var response = Cu.getWeakReference(channel).get();
    var chann = this._channels;
    for (var idx = chann.length - 1; idx > -1; idx--) {
      var request = chann[idx].get();
      if (request !== null) {
        if (request === response) {
          chann.splice(idx, 1);
          return this._logins.splice(idx, 1)[0];
        }
      } else {
        // remove request without response (from cache)?
        chann.splice(idx, 1);
        this._logins.splice(idx, 1);
      }
    }
    return null;
  },


  findSavedLogin_byurl: function(url) {
    for (var idx = this._channels.length - 1; idx > -1; idx--) {
      var request = this._channels[idx].get();
      if (request) {
        if (request.URI.spec == url) {
          console.log('findSavedLogin_byurl', url);
        }
      }
    }
  }


};


// pageshow event => call updateUI for non http/https protocols and cached pages
function updateNonNetworkDocuments(evt) {
  var doc = evt.target;
  var uri = doc.documentURIObject; // TODO pegar doc.location?
  if (isTopWindow(doc.defaultView) === false) {
    return;
  }

  if (isSupportedScheme(uri.scheme)) {
    // BUG rightclick>show image ==> evt.persisted=false
    // BUG google => br.mozdev=>back=>fw fica login do goog
    if (evt.persisted) {
      // http top doc from cache
      var tabLogin = TabLogin.getFromDomWindow(doc.defaultView);
      if (tabLogin !== null) {
        var tab = tabLogin.tabElement;
        RedirDetector.invalidateTab(tab); // invalidate redir before calling defineTopWindowLogin
        var dummy = defineTopWindowLogin(uri, tabLogin);
        updateUI(tab);
      }
    }

  } else { // about: etc
    // request listener was not called
    var tabLogin = TabLogin.getFromDomWindow(doc.defaultView);
    if (tabLogin !== null) {
      console.log('updateNonNetworkDocuments del',uri.scheme);
      var tab = tabLogin.tabElement;
      tab.removeAttribute("multifox-tab-current-tld");
      RedirDetector.resetTab(tab);
      tabLogin.setTabAsAnon();
      updateUI(tab);
    }
  }

}


const HttpHeaders = {
  visitLoop: {
    values: null,
    visitHeader: function(name, value) {
      var n = name.toLowerCase();
      if (n in this.values) {
        this.values[n] = value;
      }
    }
  },

  fromRequest: function(request) {
    var nameValues = {
      //"cookie": null, //for debug only
      "authorization": null
    }
    this.visitLoop.values = nameValues;
    request.visitRequestHeaders(this.visitLoop);
    return nameValues;
  },

  fromResponse: function(response) {
    var nameValues = {
      "set-cookie": null,
      "www-authenticate": null
    }
    this.visitLoop.values = nameValues;
    response.visitResponseHeaders(this.visitLoop);
    return nameValues;
  }
};


function getChannelWindow(channel) {
  if (channel.notificationCallbacks) {
    try {
      return channel
              .notificationCallbacks
              .getInterface(Ci.nsILoadContext)
              .associatedWindow;
    } catch (ex) {
      //util2.logEx("channel.notificationCallbacks ", channel.notificationCallbacks, channel.URI.spec, ex);
    }
  }

  if (channel.loadGroup && channel.loadGroup.notificationCallbacks) {
    try {
      return channel
              .loadGroup
              .notificationCallbacks
              .getInterface(Ci.nsILoadContext)
              .associatedWindow;
    } catch (ex) {
      util2.logEx("channel.loadGroup", channel.loadGroup, channel.URI.spec, ex);
    }
  }

  //var isChrome = context.associatedWindow instanceof Ci.nsIDOMChromeWindow;
  //return context.isContent ? context.associatedWindow : null;
  //console.log("LOAD CONTEXT FAIL " + channel.URI.spec);
  return null; // e.g. <link rel=prefetch> <link rel=next> ...
}
/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Multifox.
 *
 * The Initial Developer of the Original Code is
 * Jeferson Hultmann <hultmann@gmail.com>
 * Portions created by the Initial Developer are Copyright (C) 2011
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */


/*

BROWSING PROTOCOL


UPDATING TAB ATTRIBUTES: "multifox-tab-id-provider-tld-enc" + "multifox-tab-id-provider-user-enc"
========================================================================

*** START => SUBMIT
- in progress => user/pw submitted and copied to tab attr: "multifox-login-submit-domain" + "multifox-login-submit-user"

*** END => submitLandPageFound
- remove attrs multifox-login-submit-*
- success => commit: setLogin
- fail => restore data domain

*** REQUEST
- login in progress?
  - channel: LOAD_DOCUMENT_URI + top document (ignore subdocs)
  - POST request detected ==> onNewDoc_moveData_loginSubmitted
  - move cookies/data to a namespace


*** RESPONSE



*** DOCUMENT CREATED
*/

/*
function getChannelResponseLogin(channel, win) {
  var tabLogin = TabLogin.getFromDomWindow(win);
  if (tabLogin === null) {
    return null; // tab not found: request from chrome (favicon, updates, <link rel="next"...)
  }


  if (isWindowChannel(channel)) {
    if (tabLogin.isLoginInProgress) {
      return TabLogin.getLoginInProgress(tabLogin.tabElement);
    }

    if (isTopWindow(win)) {
      return tabLogin; // channel will replace window
    }
    return getSubElementLogin(channel.URI, tabLogin, null);

  } else {
    return getSubElementLogin(channel.URI, tabLogin, win);
  }

  if (tabLogin.isLoggedIn === false) {
    return null;
  }

}
*/


function getRequestChannelUser(channel, win) {
  var tabLogin = TabLogin.getFromDomWindow(win);
  if (tabLogin === null) {
    var chromeWin = WindowParents.getChromeWindow(win);
    if (chromeWin && (chromeWin.document.documentURI === "chrome://global/content/viewSource.xul")) {
      // view source window
      return OpenerWindowIdentity.findFromOpenerSelectedTab(win);
    }
    //console.log("getRequestChannelUser", "NO TAB", channel.URI.spec);
    return null; // tab not found: request from chrome (favicon, updates, <link rel="next"...)
  }


  var requestUri = channel.URI;
  if (isSupportedScheme(requestUri.scheme) === false) {
    // tabLogin is invalid ==> use default session
    tab.removeAttribute("multifox-tab-current-tld");
    tabLogin.setTabAsAnon();
    return null;
  }


  if (isWindowChannel(channel)) {
    // channel will create a document (top or frame)
    if (tabLogin.isLoginInProgress) { // flagged by SubmitObserver
      console.log('isLoginInProgress new doc', channel.URI.spec);
      // obs: getLoginInProgress and getFromDomWindow may both be valid (and different)
      tabLogin = TabLogin.getLoginInProgress(tabLogin.tabElement);
      onNewDoc_moveData_loginSubmitted(tabLogin); // login may be successful, so move cookies now. it may be canceled later.
      return tabLogin;
    }

    if (isTopWindow(win)) {
      // channel will replace the current top document
      return defineTopWindowLogin(requestUri, tabLogin);
    } else {
      // frames shouldn't modify tab identity (obs: frames may have a login, e.g. facebook)
      return getSubElementLogin(requestUri, tabLogin, null);
    }

  } else {
    // img, css, xhr...
    if (tabLogin.isLoginInProgress) {
      // A XHR may set a cookie in default session after a login is detected.
      // We should use login for subelements only after data is moved to a Multifox session.
      if (tabLogin.hasLoginInProgressMoveData) {
        tabLogin = TabLogin.getLoginInProgress(tabLogin.tabElement);
      }
    }
    return getSubElementLogin(requestUri, tabLogin, win);
  }
}


function defineTopWindowLogin(requestUri, prevLogin) {
  console.log("defineTopWindowLogin", requestUri.spec);
  var tab = prevLogin.tabElement;
  var redirInvalidated = tab.hasAttribute("multifox-redir-invalidate");
  if (redirInvalidated) {
    RedirDetector.resetTab(tab);
  }

  var requestTld = getTldFromHost(requestUri.host);
  var prevTopTld = null; // null => new tab or unsupported scheme
  if (tab.hasAttribute("multifox-tab-current-tld")) {
    prevTopTld = tab.getAttribute("multifox-tab-current-tld");
    if (prevTopTld === requestTld) {
      // prevLogin is still valid!
      // BUG login happens in another tab => this one will keep anonymous (LoginData.getDefaultLogin is never called)
      return prevLogin.isNewUser ? null : prevLogin;
    }
  }

  // New TLD! (prevLogin is [probably] invalid for the new top document)

  var newLogin = defineNewTldLogin(requestTld, tab, prevLogin, prevTopTld, redirInvalidated);
  if (newLogin === null) {
    console.log('requestTld',requestTld);
    prevLogin.setTabAsAnon(); // there is no login provider for requestTld
  } else {
    newLogin.saveToTab();
  }
  updateUI(tab); // show new user now (instead of waiting new dom)
  return newLogin;
}


function defineNewTldLogin(requestTld, tab, prevLogin, prevTopTld, redirInvalidated) {
  var prevPrevTopTld = tab.hasAttribute("multifox-tab-previous-tld")
                     ? tab.getAttribute("multifox-tab-previous-tld") : null;

  // we know 3 TLDs:
  //   requestTld     - new tld
  //   prevTopTld     - previous tld (associated with prevLogin)
  //   prevPrevTopTld - previous to previous tld

  tab.setAttribute("multifox-tab-previous-tld", prevTopTld);
  tab.setAttribute("multifox-tab-current-tld", requestTld);

  console.log('new tld ', prevPrevTopTld, "==>", prevTopTld, '==>', requestTld);

  // cross login in progress
  var commitCrossLogin = inheritFromPreviousTld(redirInvalidated, prevPrevTopTld, prevTopTld, requestTld, prevLogin);
  if (commitCrossLogin) {
    console.log("COMMITCROSSLOGIN", redirInvalidated, prevPrevTopTld, prevTopTld, requestTld, prevLogin);
    //tab.ownerDocument.defaultView.alert("commitCrossLogin "+ "\n"+redirInvalidated+ "\n"+ prevPrevTopTld+ "\n"+ prevTopTld+ "\n"+ requestTld+ "\n"+ prevLogin);
    copyData_fromDefault(requestTld, prevLogin); // requestTld=default
    tab.setAttribute("multifox-cross-login-commited-wait-landing-page", "true");
    tab.linkedBrowser.addEventListener("DOMContentLoaded", checkEmptyPage, false);
    return prevLogin;
  }

  // cross login done
  if (tab.hasAttribute("multifox-cross-login-commited-wait-landing-page")) {
    return prevLogin;
  }

  // popups
  if (tab.ownerDocument.documentElement.hasAttribute("multifox-window-uninitialized")) { // BUG com sessionrestore, vai processar ao abrir nova aba; considerar o no. de abas?
    tab.ownerDocument.documentElement.removeAttribute("multifox-window-uninitialized");
    // prevLogin is from opener
    var openerPrevLogin = OpenerWindowIdentity.findFromWindow(tab.linkedBrowser.contentWindow);
    if (openerPrevLogin !== null) {
      // opener tab found
      prevLogin = TabLogin.create(tab, openerPrevLogin.encodedUser, openerPrevLogin.encodedTld);
    }
  }

  // prevLogin: may keep user => different tld, same login
  return LoginData.getDefaultLogin(requestTld, prevLogin);
}



function getJsCookieLogin(domWin) { // getJsMethodLogin
  // tabLogin can be anon and the new tabLogin can be a different login (e.g. domWin=facebook iframe)
  var tabLogin = TabLogin.getFromDomWindow(domWin);
  console.assert(tabLogin !== null, "is contentDoc not from a tab?");


  if (isTopWindow(domWin)) {
    if (tabLogin.isLoginInProgress) {
      if (tabLogin.hasLoginInProgressMoveData) {
        // data already moved to new user
        tabLogin = TabLogin.getLoginInProgress(tabLogin.tabElement);
      }
    }
    return tabLogin;
  }


  // domWin=iframe


  if (tabLogin.isLoginInProgress) {
    if (tabLogin.hasLoginInProgressMoveData) {
      return TabLogin.getLoginInProgress(tabLogin.tabElement);
    } else {
      return tabLogin;
    }
  }



  var tld = getTldFromHost(domWin.document.location.hostname);
  var iframeLogin = LoginData.getDefaultLogin(tld, tabLogin);
  if (iframeLogin !== null) {
    return iframeLogin;

  } else {
    // iframe tld is not a consumer ==> third party iframe
    console.log('getJsCookieLogin - executing method from anon iframe', tld, '- top ', domWin.top.document.location);
    return tabLogin.toAnon();
  }
}


// tab element = css, img, iframe, ... ==> not called for top doc, so tabLogin is already valid
function getSubElementLogin(elemUri, tabLogin, elementWindow) {
  var tldElem = getTldFromHost(elemUri.host); // TODO enviar ja como tld

  if (elementWindow === null) {
    // elemUri is an iframe!
    var a = LoginData.getDefaultLogin(tldElem, tabLogin); // tabLogin = dummy value
    if (a !== null) {
      // frame has a known login
      console.log("getSubElementLogin iframe login found!", tldElem);
      return a;
    }

    // resource
    //console.log("getSubElementLogin anonResource", tldElem);
    return tabLogin.toAnon();
  }


  // ==> img/style/script

  if (isTopWindow(elementWindow) === false) {
    // elemUri = elem inside elementWindow (an iframe) ==> different TLD from tab?
    // elementWindow.documentURIObject may be null
    var iframe = elementWindow.location;
    var schFrame = iframe.protocol.substr(0, iframe.protocol.length - 1);
    //console.log("getSubElementLogin element/iframe", tldElem, elementWindow, schFrame, elementWindow.location, "/", elementWindow.documentURIObject === null, elementWindow.documentURIObject);
    if (isSupportedScheme(schFrame) === false) {
      return null;
    }
    var iframeLogin = LoginData.getDefaultLogin(getTldFromHost(iframe.hostname), tabLogin); // tabLogin = dummy value

    if (iframeLogin !== null) {
      if (iframeLogin.isNewUser) {
        return null;
      }
      if (LoginData.providerHasThisConsumer(stringToHex(tldElem), iframeLogin.encodedUser, iframeLogin.encodedTld)) { // WTF
        return iframeLogin;
      }
    }

    // resource
    //console.log("getSubElementLogin anonResource", tldElem);
    return tabLogin.toAnon();
  }


  // resource (css/js) from top window

  if (tabLogin.isLoggedIn === false) {
    return null;
  }

  // tldLogin = google.com
  // tab host = www.youtube.com
  // elemUri  = img.youtube.com
  if (tldElem === tabLogin.plainTld) {
    return tabLogin;
  }

  var tabTld = tabLogin.getTabTld(); //= tabLogin.tabElement.getAttribute("multifox-tab-current-tld");
  if (tldElem === tabTld) {
    return tabLogin;
  }

  // tldElem != tabLogin.plainTld ==> is tldElem logged in?

  if (LoginData.providerHasThisConsumer(stringToHex(tldElem), tabLogin.encodedUser, tabLogin.encodedTld)) {
    return tabLogin;
  }

  // resource
  return tabLogin.toAnon();
}


/*
CROSS-TLD LOGIN PROTOCOL
TLD COPY LOGIN PROTOCOL
========================
*/

function inheritFromPreviousTld(redirInvalidated, prevPrevTopTld, prevTopTld, requestTld, prevLogin) {
  if (redirInvalidated ||
     (prevPrevTopTld === null) || (prevTopTld === null) ||
     (prevLogin === null)      || (prevLogin.isLoggedIn === false)) {
    return false;
  }

  console.log('REDIR! 0');
  return (prevPrevTopTld === requestTld) &&
         (LoginData.hasConsumer(stringToHex(requestTld)) === false) &&
         LoginData.hasConsumer(stringToHex(prevTopTld));
}


function checkEmptyPage(evt) { // DOMContentLoaded
  var contentDoc = evt.target;
  console.log('checkEmptyPage top=', contentDoc.defaultView === contentDoc.defaultView.top, contentDoc.location);
  if (isEmptyPage(contentDoc)) {
    return;
  }
  var tab = WindowParents.getTabElement(contentDoc.defaultView);
  //var tabLogin = TabLogin.getFromDomWindow(contentDoc.defaultView);
  //var tab = tabLogin.tabElement;
  tab.removeAttribute("multifox-cross-login-commited-wait-landing-page");
  tab.linkedBrowser.removeEventListener("DOMContentLoaded", checkEmptyPage, false);
}


// ugly workaround, but it works most of the time
const RedirDetector = {

  onMouseDown: function(evt) {
    //console.log('onmouse', evt.currentTarget, evt.originalTarget, evt.target.ownerDocument.location, evt.altKey, evt.ctrlKey, evt.keyCode, evt.charCode, evt.keyCode !== 0);
    if (evt.button === 0) {
      var win = evt.currentTarget;
      RedirDetector._invalidateSelectedTab(win);
    }
  },

  onKeyDown: function(evt) {
    // keypress--> charCode=0
    //console.log('onkey', evt.currentTarget, evt.originalTarget, evt.target.ownerDocument.location, evt.altKey, evt.ctrlKey, evt.keyCode, evt.charCode, evt.keyCode !== 0);
    var k = Ci.nsIDOMKeyEvent;
    //if (evt.altKey || evt.ctrlKey || (evt.keyCode !== 0) || (evt.charCode === k.DOM_VK_SPACE)) {
    if (evt.keyCode < 33) {
      // keyCode ==> backspace, enter
      // DOM_VK_SPACE ==> submit form
      var win = evt.currentTarget;
      RedirDetector._invalidateSelectedTab(win);
    }
  },

  _invalidateSelectedTab: function(win) {
    if (win === null) {
      return;
    }
    var tab = WindowParents.getTabElement(win);
    /*
    var tab = null;
    var tabLogin = TabLogin.getFromDomWindow(win); // BUG win: chrome or dom?
    if (tabLogin === null) {
      if ("getBrowser" in win) {
        tab = win.getBrowser().selectedTab;
      }
    } else {
      tab = tabLogin.tabElement; // BUG se tab nao tem login, precisa invalidar?
    }*/
    RedirDetector.invalidateTab(tab);
  },

  invalidateTab: function(tab) {
    if ((tab === null) || tab.hasAttribute("multifox-redir-invalidate")) {
      return;
    }

    tab.setAttribute("multifox-redir-invalidate", "true");
  },

  resetTab: function(tab) {
    tab.removeAttribute("multifox-redir-invalidate");
  }
};


function isWindowChannel(channel) {
  return (channel.loadFlags & Ci.nsIChannel.LOAD_DOCUMENT_URI) !== 0;
}


function isTopWindow(win) {
  return win === win.top;
}


function isEmptyPage(contentDoc) {
  //var a = contentDoc.getElementsByTagName("head")[0].getElementsByTagName("*").length;
  var body = contentDoc.getElementsByTagName("body");
  var tags = body.length > 0 ? body[0].getElementsByTagName("*").length : 0;

  // TODO
  /*
  var all = contentDoc.forms;
  var qty = 0;
  for (var idx = all.length - 1; idx > -1; idx--) {
    qty += countPasswordFields(all[idx], ???);
  }
  */
  console.log("isEmptyPage", tags < 15, contentDoc.location);
  return tags < 15;
}





/*
FORM LOGIN
  SUBMIT
    Land Page
      Fail
      Success




browse ANON_TLD
[...]

browser AUTH_TLD
save multifox-cross-login-try-anon-tld=ANON_TLD
[...]

returns to ANON_TLD. Should add ANON_TLD to AUTH_TLD space?
- redir? yes!
- just made a login in AUTH_TLD? yes!
- How?
  - click on last page?
  - redir

- redir se nao houve click/key/fav na last page setAtribute("multifox-last-page-had-user-action", "") (nao contar chrome ev)



ANON_TLD_2 (logged or not)
  - set watched-tld="new_tld_1"
  - land page:
    - login form => abort() (use form protocol)
  - redirect to "new_tld_1"

ANON_TLD_1
*/

/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Multifox.
 *
 * The Initial Developer of the Original Code is
 * Jeferson Hultmann <hultmann@gmail.com>
 * Portions created by the Initial Developer are Copyright (C) 2011
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */


/*
LOGIN
=====
  post + login form: submit detected
    save username
      request will isolate data to login tld
      wait land page (it can be in another tld)
        success: commit
        fail: undo move data
*/

const SubmitObserver = {
  QueryInterface: XPCOMUtils.generateQI([Ci.nsIFormSubmitObserver]),

  notify: function (form, win, actionURI, cancelSubmit) {
    console.assert(form.ownerDocument.defaultView === win, "form.ownerDocument.defaultView != win");
    this._notify(form, win);
    return true;
  },

  _notify: function(form, win) {
    if (form.method.toUpperCase() !== "POST") {
      return;
    }

    var locWin = win.document.location;
    var schWin = locWin.protocol.substr(0, locWin.protocol.length - 1);
    //var uri = win.document.documentURIObject;
    //console.log('_notify', uri.scheme, uri.spec, isSupportedScheme(uri.scheme), '*',win.document.location.href);
    if (isSupportedScheme(schWin) === false) {
      console.log('_notify false ', isSupportedScheme(schWin));
      return;
    }

    var tab = WindowParents.getTabElement(win);
    if (tab === null) {
      return null; // chrome form?
    }
    /*
    var tabLogin = TabLogin.getFromDomWindow(win);
    if (tabLogin === null) {
      tab = WindowParents.getTabElement(win);
      if (tab === null) {
        return null; // chrome form?
      }
    } else {
      tab = tabLogin.tabElement;
    }
    */

    if (countPasswordFields(form, true) === 0) {
      return; // no password provided => not a login form
    }

    var username = findUserName(form);
    if (username === null) {
      console.log("SubmitObserver: NOP, username not found, confirm pw form?");
      return; // TODO username = "random" or error message (icon);
    }
    /*
    if (username === null) {
      if (tabLogin.isLoggedIn) {
        console.log("SubmitObserver: NOP (username not found, confirm pw form?)");
      } else {
        console.log("SubmitObserver: NOP (username not found)");
      }
      return; // TODO username = "random" or error message (icon);
    }
    console.assert(username.length > 0, "SubmitObserver: findUserName");

    var d=stringToHex(username);
    console.log('username hex', username, d, hexToString(d));
    */
    // keep tab identity, request will need to copy its cookies
    TabLogin.setLoginInProgress(username, locWin.hostname, tab);
    tab.linkedBrowser.addEventListener("DOMContentLoaded", SubmitObserver._detectSubmitLandPage, false);
  },


  _detectSubmitLandPage: function(evt) { // DOMContentLoaded
    var contentDoc = evt.target;
    if (isEmptyPage(contentDoc)) {
      return;
    }

    // land page found!
    var tab = WindowParents.getTabElement(contentDoc.defaultView);
    //var tabLogin = TabLogin.getFromDomWindow(contentDoc.defaultView);
    //var tab = tabLogin.tabElement;
    tab.linkedBrowser.removeEventListener("DOMContentLoaded", SubmitObserver._detectSubmitLandPage, false);


    var newLogin = TabLogin.getLoginInProgress(tab);
    var moveMode = TabLogin.removeLoginInProgress(tab);
    console.assert(newLogin.plainTld  !== null, "multifox-login-submit-domain");
    console.assert(newLogin.plainUser !== null, "multifox-login-submit-user");


    if (hasLoginForm(contentDoc)) {
      console.log("submitLandPageFound", "login error detected", contentDoc.documentURI);
      // cancel login
      if (moveMode === "user") {
        moveMultifoxDomainDataToAnotherUser_cancel_login(newLogin);
      } else {
        SubmitObserver._moveDomainDataToDefault_cancel_login(newLogin);
      }

    } else {
      // commit login!
      console.log("submitLandPageFound", "login successful detected", contentDoc.documentURI);
      newLogin.saveToTab();
      updateUI(tab, true);
    }
  },

  // TODO moveDomainDataTo ... ***previous domain***
  _moveDomainDataToDefault_cancel_login: function(tabLogin) { // restoreDataDomain
    var srcHost = tabLogin.formatHost(tabLogin.plainTld);
    var all = removeTldData_cookies(srcHost);

    console.log('===>_moveDomainDataToDefault_cancel_login ', tabLogin.toString(), "=", srcHost, "_", all.length);

    for (var idx = all.length - 1; idx > -1; idx--) {
      var cookie = all[idx];
      var hostData = InternalHost.parseHost(cookie.host);
      if (hostData !== null) {
        copyCookieToNewHost(cookie, hostData.realHost);
      }
    }

    var all2 = removeTldData_LS(srcHost);

  }

};


function moveMultifoxDomainDataToAnotherUser_cancel_login(tabLogin) { // TODO
}



function copyDataToAnotherUser(tabTld, newLogin, prevLogin) {
  console.assert(prevLogin.plainTld === newLogin.plainTld, "copyDataToAnotherUser tld");
  if (prevLogin.equals(newLogin)) {
    return; // same user, do nothing
  }

  var tld = prevLogin.formatHost(tabTld);
  // don't remove data from current user, it may be still valid
  // some cookies may be unrelated to this login
  var all = getAllCookiesFromHost(tld);
  //var all = removeTldData_cookies(tld);

  console.log("copyDataToAnotherUser", tabTld, all.length, "cookies.", prevLogin.toString(), newLogin.toString());
  var cookie;
  var hostData;
  for (var idx = all.length - 1; idx > -1; idx--) {
    cookie = all[idx];
    hostData = InternalHost.parseHost(cookie.host);
    if (hostData !== null) {
      copyCookieToNewHost(cookie, newLogin.formatHost(hostData.realHost));
    }
  }

  //var all2 = removeTldData_LS(tld);
}


// isolate cookies from domain
function copyData_fromDefault(domain, tabLogin) { // BUG se tabLogin.plainUser="" vai pensar NewAccount
  var all = getAllCookiesFromHost(domain);
  //var all = removeTldData_cookies(domain);

  console.log('===>copyData_fromDefault ', domain, tabLogin.toString(), "cookies:", all.length);
  var cookie;
  for (var idx = all.length - 1; idx > -1; idx--) {
    cookie = all[idx];
    //console.trace(toInternalHost(cookie.host, tabLogin) + cookie.host);
    console.log('->copyData_fromDefault ', cookie.host, tabLogin, tabLogin.formatHost(cookie.host));
    copyCookieToNewHost(cookie, tabLogin.formatHost(cookie.host));
  }

  console.log('copyData_fromDefault 2');
  var all2 = removeTldData_LS(domain);
  console.log('/===>copyData_fromDefault ');
}


function hasLoginForm(doc) { // TODO check frames
  if ("forms" in doc === false) {
    return false; // xml
  }

  var all = doc.forms;
  for (var idx = all.length - 1; idx > -1; idx--) {
    var qty = countPasswordFields(all[idx], false);
    if (qty > 0) {
      return true;
    }
  }

  return false;
}


function countPasswordFields(form, populatedOnly) {
  var qty = 0;
  var all = form.elements;
  var INPUT = Ci.nsIDOMHTMLInputElement;
  for (var idx = all.length - 1; idx > -1; idx--) {
    var elem = all[idx];
    if ((elem instanceof INPUT) && (elem.type === "password")) {
      if (isElementVisible(elem)) {
        if (populatedOnly && (elem.value.trim().length === 0)) {
          continue;
        }
        qty++;
      }
    }
  }
  return qty;
}


function findUserName(form) {
  console.log("findUserName");
  var INPUT = Ci.nsIDOMHTMLInputElement;
  var lastTextField = null;
  var all = form.elements;
  var elem;

  for (var idx = 0, len = all.length; idx < len; idx++) {
    elem = all[idx];
    if ((elem instanceof INPUT) === false) {
      continue;
    }
    switch (elem.type) {
      case "text":
      case "email":
      case "url":
      case "tel":
      case "number":
      case "password":
        break;
      default:
        continue;
    }
    if ((elem.value.trim().length === 0) || isElementVisible(elem) === false) {
      // ignore empty/hidden fields
      console.log("findUserName", "element ignored", elem.name);
      continue;
    }
    if (elem.type === "password") {
      if (lastTextField !== null) {
        return lastTextField;
      }
    } else {
      lastTextField = elem.value;
      console.log("findUserName", "found", lastTextField);
    }
  }
  return lastTextField;
}


function isElementVisible(elem) {
  // some elements with display:none/visibility:hidden are already removed from form
  var win = elem.ownerDocument.defaultView;
  var val = win.getComputedStyle(elem, "").getPropertyValue("visibility");
  console.assert(val === "visible", "visibility", val);
  console.log('isElementVisible', "getBoundingClientRect().width " + elem.getBoundingClientRect().width, typeof elem.getBoundingClientRect().width, elem.tagName, elem.type, " name:" + elem.name);// + "=" + elem.value);

  return elem.getBoundingClientRect().width > 0;
  //return elem.value.trim().length > 0;
}


function onNewDoc_moveData_loginSubmitted(newLogin) {
  if (newLogin.hasLoginInProgressMoveData) {
    return;// newLogin;
  }

  console.log("onNewDoc_moveData_loginSubmitted");
  var currentLogin = new TabInfo(newLogin.tabElement); // TODO currentLogin is tabLogin?
  currentLogin.setTabAsAnon();

  // new login from a logged in tab
  if (currentLogin.isLoggedIn) {
    // tab is marked as logged in, but it is currently logged out (there is a login form...)
    currentLogin.setLoginInProgressMoveData("user");
    console.log("copyDataToAnotherUser logged=", currentLogin.toString(), newLogin.toString());
    copyDataToAnotherUser(newLogin.plainTld, newLogin, currentLogin);

  // "New Account" tab
  } else if (currentLogin.isNewUser) {
    currentLogin.setLoginInProgressMoveData("default");
    console.log("copyData_fromDefault logged=NewAccount");
    copyData_fromDefault(newLogin.plainTld, newLogin);

  // anonymous tab
  } else {
    currentLogin.setLoginInProgressMoveData("default");
    console.log("copyData_fromDefault logged=anon");
    copyData_fromDefault(newLogin.plainTld, newLogin);
  }

  //return newLogin;
}
/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Multifox.
 *
 * The Initial Developer of the Original Code is
 * Jeferson Hultmann <hultmann@gmail.com>
 * Portions created by the Initial Developer are Copyright (C) 2011
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */


/*

this._counterTld= {
  "abcdefghj-iabc-abc": 5
}

this._providers = {
  "google.com": [
    "efghijklmnop",
    "abcdefghijkl"
  ],
  "twtter.com": [
    "efghijklmnop",
  ]


this._consumers = {
  "my-app.com": [
    "google.com",
    "facebook.com",
    "twitter.com"
  ],
  "orkut.com": [
    "google.com"
  ],
*/

const LoginData = {
  _providers: null,
  _consumers: null,
  _counterTld: null, // trigger _update
  _invalidated: true,

  _reset: function() {
    this._providers = { __proto__: null };
    this._consumers = { __proto__: null };
    this._counterTld = { __proto__: null };
    this._invalidated = false;
  },

  shutdown: function() {
    this._providers = null;
    this._consumers = null;
    this._counterTld = null;
  },

  _invalidate: function() {
    console.trace("_invalidate");
    this._invalidated = true;
  },

  _ensureValid: function() {
    if (this._invalidated) {
      this._update();
      this._invalidated = false;
    }
  },


  setTabAsDefaultLogin: function(tab) {
    var tabLogin = new TabInfo(tab);
    if (tabLogin.hasUser) {
      this.setDefaultLogin(tabLogin.getTabTldEncoded(), tabLogin.encodedUser, tabLogin.encodedTld);
    }
  },

  //setDefaultLoginForTab vs. setDefaultLoginForResource
  setDefaultLogin: function setDefaultLogin_(consumerTldEnc, providerUserEnc, providerTldEnc) {
    console.assert(typeof consumerTldEnc === "string", "consumerTldEnc=" + consumerTldEnc);
    console.assert(typeof providerUserEnc === "string",   "providerUserEnc=" + providerUserEnc);
    console.assert(typeof providerTldEnc === "string",    "providerTldEnc=" + providerTldEnc);
    //console.log("LoginData.setDefaultLogin", consumerTldEnc, providerUserEnc, providerTldEnc, hexToString(consumerTldEnc), hexToString(providerUserEnc), hexToString(providerTldEnc));

    this._ensureValid();

    if ((consumerTldEnc in this._consumers) === false) {
      console.log("LoginData.setDefaultLogin", "consumerTldEnc not found", consumerTldEnc);
      return;
    }

    if ((providerTldEnc in this._providers) === false) {
      console.log("LoginData.setDefaultLogin", "providerTldEnc not found", providerTldEnc, hexToString(providerTldEnc));
      return;
    }

    var consumerProviders = this._consumers[consumerTldEnc]; // usually len=1
    var idx2 = consumerProviders.indexOf(providerTldEnc);
    if (idx2 > 0) {                 // [0] => already the default
      consumerProviders.splice(idx, 1);
      consumerProviders.unshift(consumerTldEnc); // BUG consumerTldEnc ou providerTldEnc???
      console.log("LoginData.setDefaultLogin updated0", providerTldEnc, consumerTldEnc);
    }

    var allUsers = this._providers[providerTldEnc];
    var idx = allUsers.indexOf(providerUserEnc);
    switch (idx) {
      case 0:
        // nop: user is already the default
        break;
      case -1:
        // user not found (cookies cleared?)
        if (providerUserEnc === TabLogin.NewAccount) {
          console.log("LoginData.setDefaultLogin NewAccount", providerTldEnc, this._providers[providerTldEnc].toSource(), allUsers.toSource());
          console.trace();
          console.log('_update\nProviders:' + JSON.stringify(this._consumers, null, 2)
               + "\n==========\nConsumers:" + JSON.stringify(this._providers, null, 2)
                   + "\n==========\nHosts:" + JSON.stringify(this._counterTld, null, 2));


          allUsers.unshift(TabLogin.NewAccount);
        }
        break;
      default:
        allUsers.splice(idx, 1);
        allUsers.unshift(providerUserEnc); // [0] => default
        console.log("LoginData.setDefaultLogin updated1", hexToString(providerUserEnc), providerUserEnc, providerTldEnc, this._providers[providerTldEnc].toSource());
        break;
    }
  },

/*
  reuseTabInfo: function(tld, fallbackTabInfo) {
  getDefaultLogin: function(tld, tab) {

*/
  getDefaultLogin: function(tld, fallbackTabInfo) {
    this._ensureValid();

    var encTld = stringToHex(tld);
    if ((encTld in this._consumers) === false) {
      if (hexToString(encTld).indexOf('.faceboo') > -1) {
        console.trace(tld);
      }
      console.log("LoginData.getDefaultLogin - not a consumer", hexToString(encTld));
      return null;
    }

    // obs: loginUser16/loginTld may be null
    var loginTlds = this._consumers[encTld];
    var defaultFormTld = loginTlds[0];
    if (defaultFormTld === fallbackTabInfo.encodedTld) {
      //console.log("LoginData.getDefaultLogin", "don't change user", fallbackTabInfo.plainUser, fallbackTabInfo.plainTld);
      return fallbackTabInfo; // reuse/don't change current user
    }

    console.assert(defaultFormTld in this._providers, "this._providers " + defaultFormTld + "/" + tld);
    var users = this._providers[defaultFormTld];
    console.assert(users.length > 0, "users.length");
    var defaultUser = users[0];


    var newLogin = TabLogin.create(fallbackTabInfo.tabElement, defaultUser, defaultFormTld);

    /*
    if (defaultUser === TabLogin.NewAccount) {
      users.splice(0, 1); // workaround - middle click on "New Account" -- BUG next calls won't be NewAccount
      console.log("LoginData.getDefaultLogin", "user NEW ", defaultFormTld, fallbackTabInfo.toString());
    }
    */

    //console.log('_update\n' + JSON.stringify(this._consumers, null, 2) + "\n==========\n" + JSON.stringify(this._providers, null, 2));
    console.log("LoginData.getDefaultLogin0", "user found ", defaultUser, defaultFormTld);
    console.log("LoginData.getDefaultLogin1", "user found ", hexToString(defaultUser), hexToString(defaultFormTld));
    return newLogin;
  },


  addCookie: function(rawHost) {
    if (InternalHost.isInternalLoggedIn(rawHost) === false) {
      return;
    }

    var hostData = InternalHost.parseHost_enc(rawHost);
    var encStrip  = hostData.enc_data;
    if (encStrip in this._counterTld) {
      this._counterTld[encStrip]++;
    } else {
      this._counterTld[encStrip] = 1;
      this._invalidate(); // new login?
      console.log("cookie ADD _invalidate!", hostData.enc_realHostTld, hostData.enc_myUser);
    }

    //if (this.providerHasThisConsumer(hostData.enc_realHostTld, hostData.enc_myUser, hostData.enc_myTld) === false) {
    //  this._invalidate(); // login added
    //}
  },


  deleteCookie: function(rawHost) {
    if (InternalHost.isInternalLoggedIn(rawHost) === false) {
      return false;
    }
    var counter = this._counterTld;
    if (counter === null) {
      return false;
    }
    var hostData = InternalHost.parseHost_enc(rawHost);
    var encStrip = hostData.enc_data;
    console.assert(encStrip in counter, "!encStrip in this._counterTld " + encStrip);
    counter[encStrip]--;
    if (counter[encStrip] > 0) {
      return false;
    }
    delete counter[encStrip];
    this._invalidate(); // login removed?
    console.log('deleteCookie invalidated', rawHost, counter[encStrip], JSON.stringify(counter, null, 2));
    return true;
  },


  _update: function _update_() {
    console.trace("LoginData._update");
    var oldProviders = this._providers; // save current defaults
    var oldConsumers = this._consumers;
    this._reset();

    var hostData;
    var users;

    var all = InternalHost.getLoggedCookies();
    setWelcomeMode(all.length === 0);

    for (var idx = all.length - 1; idx > -1; idx--) {
      var rawHost = all[idx].host;

      hostData = InternalHost.parseHost_enc(rawHost);
      if (hostData === null) {
        continue;
      }

      var encStrip  = hostData.enc_data;
      var siteTld   = hostData.enc_realHostTld; // hostConsumer
      var loginUser = hostData.enc_myUser; // encUsrProvider
      var loginTld  = hostData.enc_myTld; //  encTldProvider

      if (encStrip in this._counterTld) {
        this._counterTld[encStrip]++;
      } else {
        this._counterTld[encStrip] = 1;
      }

      if (loginTld in this._providers) {
        var aa = this._providers[loginTld];
        if (aa.indexOf(loginUser) === -1) {
          aa.push(loginUser);
        }
      } else {
        this._providers[loginTld] = [loginUser];
      }


      if (siteTld in this._consumers) {
        var bb = this._consumers[siteTld];
        if (bb.indexOf(loginTld) === -1) {
          bb.push(loginTld);
        }
      } else {
        this._consumers[siteTld] = [loginTld];
      }

    }

    // keep default users
    for (var tabTld in oldConsumers) {
      var tld = oldConsumers[tabTld][0];
      var user = oldProviders[tld][0];
      this.setDefaultLogin(tabTld, user, tld);
    }

    console.log('_update\nProviders:' + JSON.stringify(this._consumers, null, 2)
         + "\n==========\nConsumers:" + JSON.stringify(this._providers, null, 2)
             + "\n==========\nHosts:" + JSON.stringify(this._counterTld, null, 2));
  },


  providerHasThisConsumer: function(consumerTldEncoded, providerUserEncoded, providerTldEncoded) {
    this._ensureValid();
    if (consumerTldEncoded in this._consumers) {
      var providers = this._consumers[consumerTldEncoded];
      var useProviderTld = providers.indexOf(providerTldEncoded) > -1;
      if (useProviderTld) {
        if (providerTldEncoded in this._providers) {
          if (this._providers[providerTldEncoded].indexOf(providerUserEncoded) > -1) {
            return true;
          }
        }
      }
    }
    return false;
  },

  hasConsumer: function(consumerTldEncoded) {
    this._ensureValid();
    return consumerTldEncoded in this._consumers;
  },

  getEncodedTldUsers: function(enc_tabTld) { // tabTld = "orkut.com.br" (login: "XYZ.google.com")
    this._ensureValid();
    if ((enc_tabTld in this._consumers) === false) {
      return [];
    }
    var tlds = this._consumers[enc_tabTld];

    // TODO test multiple sites eg bugzilla + amo
    var tabUsers = [];
    for (var idx1 = tlds.length - 1; idx1 > -1; idx1--) {
      var loginTld = tlds[idx1];
      var users = this._providers[loginTld];
      for (var idx2 = users.length - 1; idx2 > -1; idx2--) {
        var myUser = users[idx2];
        if (myUser === TabLogin.NewAccount) {
          continue;
        }
        tabUsers.push({ username5: myUser,
                        usernamePlain: hexToString(myUser),
                        tld: loginTld
        });
      }
    }

    // alphabetical sort
    tabUsers.sort(function(userA, userB) {
      return userB.usernamePlain.localeCompare(userA.usernamePlain);
    });

    return tabUsers;
  },


  onCookieRejected: {
    QueryInterface: XPCOMUtils.generateQI([Ci.nsIObserver]),
    observe: function(subject, topic, data) {
      console.log("cookie-rejected:");
      console.log("cookie-rejected\n", subject, "\n", topic, "\n", data, "\n", subject.QueryInterface(Ci.nsIURI).spec);
    }
  },

  onCookieChanged: { // onCookieChangedInvalidateLoginData
    QueryInterface: XPCOMUtils.generateQI([Ci.nsIObserver]),
    observe: function(subject, topic, data) {
      if (LoginData._consumers === null) {
        return;
      }
      switch (data) {
        case "changed":
          //var cookie = subject.QueryInterface(Ci.nsICookie2);
          break;
        case "added":
          var cookie = subject.QueryInterface(Ci.nsICookie2);
          //console.log("cookie ADD", cookie.host, cookie.path, cookie.name, cookie.value);
          LoginData.addCookie(cookie.host);
          break;
        case "deleted":
          var cookie = subject.QueryInterface(Ci.nsICookie2);
          //console.trace("deleted "+cookie.host + "///"+cookie.name);
          LoginData.deleteCookie(cookie.host);
          break;

        case "batch-deleted":
          console.trace("batch-deleted");
          var all = subject.QueryInterface(Ci.nsIArray).enumerate();
          console.log("cookie BATCH-DELETED! " + data + all);
          while (all.hasMoreElements()) {
            var cookie = all.getNext().QueryInterface(Ci.nsICookie2);
            if (LoginData.deleteCookie(cookie.host)) {
              break; // invalidated, it is not necessary yo continue
            }
          }
          break;
        case "cleared":
          LoginData._reset();
          break;
        case "reload":
          LoginData._invalidate();
          break;
        default:
          LoginData._invalidate(); // ???
          break;
      }
    }
  }
};


/*
function checkExpBug(cookie) {
  console.log("cookie", cookie.name, new Date().getTime(), cookie.expiry, cookie.expires, "\n",
               cookie.host, cookie.path, cookie.value, cookie.rawHost, "\n",
               cookie.isDomain, cookie.isHttpOnly, new Date(1000*cookie.expiry), new Date(1000*cookie.expires), '//', new Date(cookie.creationTime), new Date(cookie.lastAccessed/1000), cookie.isSession, cookie.isSecure, cookie.policy, cookie.status);

  switch (cookie.name) {
    case "auth_token"://twttr
    case "datr"://fb
    case "GAPS"://goog
    case "PREF"://goog
    case "put_1430"://statc
    case "au"://statc
    case "__qca"://statc
    case "is_unique":
    case "is_unique_1":
    //case "twll":
    //case "_twitter_sess":
    //case "secure_session":
      break;
    default:
      return;
  }

  var hostData5 = getLoginFromInternalHost(cookie.host);
  var login5 = splitUserNameAndTld(hostData5.fullLogin);


  var h = "${INTERNAL__DOMAIN__PREFIX}." + hostData5.fullLogin + ".${INTERNAL__DOMAIN__ORIGIN_LOGIN}.${INTERNAL__DOMAIN__SUFFIX}";
  var all = getAllCookiesFromHost(h);
  console.log("checkExpBug USER", hexToString(login5[0]), "remain:", all.length, h);
  if (all.length === 0) {
    //var browserWin = Cc["@mozilla.org/appshell/window-mediator;1"].getService(Ci.nsIWindowMediator).getMostRecentWindow("navigator:browser");
    //browserWin.alert("ZERO "+cookie.host);
  }

  var browserWin = Cc["@mozilla.org/appshell/window-mediator;1"].getService(Ci.nsIWindowMediator).getMostRecentWindow("navigator:browser");
  browserWin.alert(cookie.name+" "+cookie.host + " * " + hexToString(login5[0]));

}

function splitUserNameAndTld(usrTld) {
  var idx = usrTld.indexOf(".");
  return [usrTld.substr(0, idx), usrTld.substr(idx + 1)];
}
*/
/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Multifox.
 *
 * The Initial Developer of the Original Code is
 * Jeferson Hultmann <hultmann@gmail.com>
 * Portions created by the Initial Developer are Copyright (C) 2009
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */


const PREF_COOKIE_BEHAVIOR = "network.cookie.cookieBehavior";

const Cookies = {
  _service: null,
  _prefs: null,

  start: function() {
    this._service = Cc["@mozilla.org/cookieService;1"].getService().QueryInterface(Ci.nsICookieService);
    this._prefs = Cc["@mozilla.org/preferences-service;1"].getService(Ci.nsIPrefBranch2);
    this._prefListener.behavior = this._prefs.getIntPref(PREF_COOKIE_BEHAVIOR);
    this._prefs.addObserver(PREF_COOKIE_BEHAVIOR, this._prefListener, false);
  },

  stop: function() {
    this._service = null;
    this._prefs.removeObserver(PREF_COOKIE_BEHAVIOR, this._prefListener);
    this._prefs = null;
  },


  _prefListener: {
    behavior: -1,

    QueryInterface: XPCOMUtils.generateQI([Ci.nsIObserver]),
    // aTopic=nsPref:changed aData=network.cookie.cookieBehavior
    observe: function(aSubject, aTopic, aData) {
      this.behavior = aSubject
                      .QueryInterface(Ci.nsIPrefBranch)
                      .getIntPref(PREF_COOKIE_BEHAVIOR);
      console.log("pref! " + aSubject + aTopic + aData + this.behavior);
    }

  },

  setCookie: function(tabLogin, originalUri, originalCookie, fromJs) {
    var uri = tabLogin.formatUri(originalUri);
    var val = convertCookieDomain(originalCookie, tabLogin);

    if (this._prefListener.behavior === 0) {
      this._setCookie(fromJs, uri, val);
      return;
    }

    var p = this._prefs;
    p.removeObserver(PREF_COOKIE_BEHAVIOR, this._prefListener);
    p.setIntPref(PREF_COOKIE_BEHAVIOR, 0);
    this._setCookie(fromJs, uri, val);
    p.setIntPref(PREF_COOKIE_BEHAVIOR, this._prefListener.behavior);
    p.addObserver(PREF_COOKIE_BEHAVIOR, this._prefListener, false);
  },

  getCookie: function(fromJs, originalUri, tabLogin) {
    var uri = tabLogin.formatUri(originalUri);

    if (this._prefListener.behavior === 0) {
      return this._getCookie(fromJs, uri);
    }

    var p = this._prefs;
    p.removeObserver(PREF_COOKIE_BEHAVIOR, this._prefListener);
    p.setIntPref(PREF_COOKIE_BEHAVIOR, 0);
    var cookie = this._getCookie(fromJs, uri);
    p.setIntPref(PREF_COOKIE_BEHAVIOR, this._prefListener.behavior);
    p.addObserver(PREF_COOKIE_BEHAVIOR, this._prefListener, false);
    return cookie;
  },


  _setCookie: function(fromJs, uri, val) {
    //console.trace('_setCookie \n'+ fromJs+ ' \n'+uri.host);//+ ' \n' + val);
    if (fromJs) {
      this._service.setCookieString(uri,
                                    null,
                                    val,
                                    null);
    } else {
      //setCookieString doesn't work for httponly cookies
      this._service.setCookieStringFromHttp(uri,   // aURI
                                            null,  // aFirstURI
                                            null,  // aPrompt
                                            val,   // aCookie
                                            null,  // aServerTime
                                            null); // aChannel
    }
  },

  _getCookie: function(fromJs, uri) {
    if (fromJs) {
      return this._service.getCookieString(uri,
                                           null);
    } else {
      return this._service.getCookieStringFromHttp(uri,   // aURI
                                                   null,  // aFirstURI
                                                   null); // aChannel
    }
  }
};


function convertCookieDomain(cookieHeader, tabLogin) {
  var objCookies = new SetCookieParser(cookieHeader, true);
  var len = objCookies.length;
  var newCookies = new Array(len);

  for (var idx = 0; idx < len; idx++) {
    var myCookie = objCookies.getCookieByIndex(idx);
    var realDomain = myCookie.getStringProperty("domain");
    if (realDomain.length > 0) {
      myCookie.defineMeta("domain", tabLogin.formatHost(realDomain));
      newCookies[idx] = myCookie.toHeaderLine();
    } else {
      newCookies[idx] = myCookie.raw;
    }
  }

  return newCookies.join("\n");//objCookies.toHeader();
}


function toLines(txt) {
  return txt.split(/\r\n|\r|\n/);
}


function SetCookieParser(cookieHeader, isSetCookie) { // TODO remove isSetCookie
  this.m_hasMeta = isSetCookie;
  this._allCookies = [];
  if (cookieHeader !== null) {
    var lines = toLines(cookieHeader);
    var len = lines.length;
    if (isSetCookie) {
      for (var idx = 0; idx < len; idx++) {
        this.parseLineSetCookie(lines[idx]);
      }
    } else {
      for (var idx = 0; idx < len; idx++) {
        this.parseLineCookie(lines[idx]);
      }
    }
  }
}

SetCookieParser.prototype = {
  parseLineSetCookie: function(headerLine) {
    var unit = new CookieUnit(headerLine);
    var items = headerLine.split(";");

    for (var idx = 0, len = items.length; idx < len; idx++) {
      var pair = splitValueName(items[idx]);
      var name = pair[0];
      var value = pair[1]; // null ==> name=HttpOnly, secure etc

      if (idx === 0) {
        if (name.length > 0 && value !== null) {
          unit.defineValue(name, value);
        } else {
          util2.logEx("_allCookies invalid", name, value, headerLine);
          break;
        }
      } else {
        unit.defineMeta(name, value);
      }
    }

    this._allCookies.push(unit);
  },

  parseLineCookie: function(headerLine) {
    var items = headerLine.split(";");
    for (var idx = 0, len = items.length; idx < len; idx++) {
      var unit = CookieUnit(items[idx]);
      var pair = splitValueName(items[idx]);
      unit.defineValue(pair[0], pair[1]);
      this._allCookies.push(unit);
    }
  },

  /*
  toHeader: function() {
    var allCookies = this._allCookies;
    var len = allCookies.length;
    //var buf = [];
    var buf = new Array(len);
    for (var idx = 0; idx < len; idx++) {
      //if (allCookies[idx].value !== null) {
      buf[idx] = allCookies[idx].toHeaderLine();
      //}
    }
    return this.m_hasMeta ? buf.join("\n") : buf.join(";");
  },

  getCookie: function(name) {
    var aCookie = this._allCookies;
    for (var idx = 0, len = aCookie.length; idx < len; idx++) {
      if (name === aCookie[idx].name) {
        return aCookie[idx];
      }
    }
    return null;
  },

  forEach: function(fn) {
    var c = this._allCookies;
    for (var idx = 0, len = c.length; idx < len; idx++) {
      fn(c[idx]);
    }
  }
  */

  get length() {
    return this._allCookies.length;
  },

  getCookieByIndex: function(idx) {
    return this._allCookies[idx];
  }
};

function CookieUnit(line) {
  this._data = {//instance value
    "_src": line
  };
}

CookieUnit.prototype = {
  _data: null,

  clone: function() {
    var c = new CookieUnit();
    for (var n in this._data) {
      c._data[n] = this._data[n];
    }
    return c;
  },

  get name() {
    var rv = this._data["_name"];
    return rv ? rv : "";
  },

  get value() {
    var rv = this._data["_value"];
    return rv ? rv : "";
  },

  get raw() {
    return this._data["_src"];
  },

  defineValue: function(name, val) {
    this._data["_name"] = name;
    this._data["_value"] = val;
  },

  //"secure":
  //"httponly":
  hasBooleanProperty: function(name) {
    //name = name.toLowerCase();
    return name in this._data;
  },

  setBooleanProperty: function(name, def) {
    if (def) {
      this._data[name] = null;
    } else {
      delete this._data[name];
    }
  },

  //"expires":
  //"domain":
  //"path":
  getStringProperty: function(name) {
    var rv = this._data[name];
    return rv ? rv : "";
    //return rv || "";
  },

  defineMeta: function(name, val) {
    name = name.toLowerCase();
    switch (name) {
      case "expires":
      case "domain":
      case "path":
      case "secure":
      case "httponly":
        this._data[name] = val;
        break;
    }
  },

  toHeaderLine: function() {//toString()
    var buf = [this.name + "=" + this.value];
    var props;

    props = ["secure", "httponly"];
    for (var idx = 0, len = props.length; idx < len; idx++) {
      var propName = props[idx];
      if (this.hasBooleanProperty(propName)) {
        buf.push(propName.toUpperCase());
      }
    }

    props = ["expires", "path", "domain"];
    for (var idx = 0, len = props.length; idx < len; idx++) {
      var propName = props[idx];
      var val = this.getStringProperty(propName);
      if (val.length > 0) {
        buf.push(propName.toUpperCase() + "=" + val);
      }
    }

    return buf.join(";");
  }
};


function splitValueName(cookie) {
  var idx = cookie.indexOf("=");
  if (idx === -1) {
    return [cookie.trim(), null];
  }

  // "a=bcd=e".split("=",2) returns [a,bcd]
  //   "abcde".split("=",2) returns [abcde]


  // MY =
  // 012^-----idx=3 length=4

  // MY =a:1=6
  // 012^-----idx=3 length=9

  var pair = ["", ""];
  pair[0] = cookie.substring(0, idx).trim();
  idx++;
  if (idx < cookie.length) {
    pair[1] = cookie.substring(idx);
  }

  return pair;
}
/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Multifox.
 *
 * The Initial Developer of the Original Code is
 * Jeferson Hultmann <hultmann@gmail.com>
 * Portions created by the Initial Developer are Copyright (C) 2011
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */




function removeTldData_LS(tld) {
  // TODO del localStorage
}



function copyCookieToNewHost(cookie, newHost) {
  var expiryTime = cookie.isSession ? 4611686018427388 : cookie.expiry; // Math.pow(2, 62)=4611686018427388000
  var mgr = Cc["@mozilla.org/cookiemanager;1"].getService(Ci.nsICookieManager2);
  mgr.add(newHost,         cookie.path,
          cookie.name,     cookie.value,
          cookie.isSecure, cookie.isHttpOnly, cookie.isSession, expiryTime);

  //console.log("/copyCookieToNewHost exp", newHost, cookie.name, cookie.isSession, cookie.expiry, expiryTime, new Date(1000 * expiryTime));
}


function removeTldData_cookies(tld) { // TODO del ".multifox-anon-1" + ".multifox-auth-1"
  var all = getAllCookiesFromHost(tld);
  console.log('removeCookies n=', all.length, tld);

  var mgr = Cc["@mozilla.org/cookiemanager;1"].getService(Ci.nsICookieManager2);
  var cookie;
  for (var idx = all.length - 1; idx > -1; idx--) {
    cookie = all[idx];
    mgr.remove(cookie.host, cookie.name, cookie.path, false);
  }
  return all;
}


function getAllCookiesFromHost(h) {
  var rv = [];
  var mgr = Cc["@mozilla.org/cookiemanager;1"].getService(Ci.nsICookieManager2);
  var COOKIE = Ci.nsICookie2;

  var _qty = 0;
  var _t = new Date().getTime();

  // TODO use mgr.getCookiesFromHost when it doesn't have bugs
  var all = mgr.enumerator;
  while (all.hasMoreElements()) {
    var cookie = all.getNext().QueryInterface(COOKIE);
    _qty++;
    if (hasRootDomain(h, cookie.host)) {
      rv.push(cookie);
    }
  }

  console.log("getAllCookiesFromHost time=" + (new Date().getTime() - _t) + "ms -- len parsed=" + _qty + " -- len rv=" + rv.length);
  return rv;
}


//getLoginFromInternalHost 1 .google.com.x-tab.x-session.jrklas.x-tab.x-session.google.com.x-tab.x-session4 - false
/*
multifox[04:06:42.337] getLoginFromInternalHost    .google.com.x-tab.x-session.jrklas.x-tab.x-session.google.com.x-tab.x-session
 [".google.com",
  ".jrklas",
  ".google.com",
  ""
]

multifox[04:06:42.337] getLoginFromInternalHost .statcounter.com.x-tab.x-session.gihfgmhegngbgogo.statcounter.com.x-tab.x-session
 [".statcounter.com",
  ".gihfgmhegngbgogo.statcounter.com",
  ""
]

*/


function hasRootDomain(domain, host) {
  if (host === domain) {
    return true;
  }

  var idx = host.lastIndexOf(domain);
  if (idx === -1) {
    return false;
  }

  //return (index == (this.length - aDomain.length)) && (prevChar == "." || prevChar == "/");
  if ((host.length - domain.length) === idx) {
    return host[idx - 1] === ".";// || prev === "/";
  }

  return false;
}
/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Multifox.
 *
 * The Initial Developer of the Original Code is
 * Jeferson Hultmann <hultmann@gmail.com>
 * Portions created by the Initial Developer are Copyright (C) 2010
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

function windowLocalStorage(obj, contentDoc) {
  if (contentDoc.documentURI === "about:blank") { // BUG 1.9.2
    console.log("windowLocalStorage doc=", contentDoc.documentURI, "/", contentDoc.location);
    return "";
  }

  //var tabLogin = TabLogin.getFromDomWindow(contentDoc.defaultView);
  //console.assert(tabLogin !== null, "is contentDoc not from a tab?");

  var tabLogin = getJsCookieLogin(contentDoc.defaultView);

  var uri = tabLogin.formatUri(contentDoc.documentURIObject);

  var principal = Cc["@mozilla.org/scriptsecuritymanager;1"]
                  .getService(Ci.nsIScriptSecurityManager)
                  .getCodebasePrincipal(uri);

  var storage = Cc["@mozilla.org/dom/storagemanager;1"]
                .getService(Ci.nsIDOMStorageManager)
                .getLocalStorageForPrincipal(principal, "");

  var rv = undefined;
  switch (obj.cmd) {
    case "clear":
      storage.clear();
      break;
    case "removeItem":
      storage.removeItem(obj.key);
      break;
    case "setItem":
      storage.setItem(obj.key, obj.val); // BUG it's ignoring https
      break;
    case "getItem":
      rv = storage.getItem(obj.key);
      break;
    case "key":
      rv = storage.key(obj.index);
      break;
    case "length":
      rv = storage.length;
      break;
    default:
      throw new Error("localStorage interface unknown: " + obj.cmd);
  }

  //console.log("localStorage " + uri.spec + "\n"+JSON.stringify(obj, null, 2) + "\n=====\nreturn " + rv);
  return rv;
}




/*
const DomApi = {

  cookie: function() {

  },
  localStorage:

};
*/
/*
multifox[02:23:12.196] documentCookie _ __utma=254116750.112059089.1311479066.1311479066.1311479066.1; __utmz=254116750.1311479066.1.1.utmccn=(direct)|utmcsr=(direct)|utmcmd=(none); OSN=ID=F9WhqZm8iWI=:S=1F2IwB1IboiQ6sO0:; lobo=ut=1311484527180:tl=0:ct=0:; TZ=180; frame=C=CommMsgs%3Fcmm%3D698395%26tid%3D5629481869631749514 string [object Object] http://www.orkut.com.br/Main#CommMsgs?cmm=698395&tid=5628823300692608797
*/
function documentCookie(obj, contentDoc) {
  switch (obj.cmd) {
    case "set":
      documentCookieSetter(obj, contentDoc);
      return undefined;
    case "get":
      var xx=99;
      try{
      xx=documentCookieGetter(obj, contentDoc);
      return xx;
      }catch(ex){
        console.trace(ex);
      } finally {
        //console.log('documentCookie _',typeof xx,xx,obj, contentDoc.location);
      }
    default:
      throw new Error("documentCookie " + obj.cmd);
  }
}


function documentCookieSetter(obj, contentDoc) {
  //var tabLogin = TabLogin.getFromDomWindow(contentDoc.defaultView);
  //console.assert(tabLogin !== null, "not a tab document?");
  //if (tabLogin.isLoginInProgress) {
  //  tabLogin = TabLogin.getLoginInProgress(tabLogin.tabElement);
  //}

  // tabLogin can be anon and the new tabLogin can be a different login (e.g. facebook iframe)
  var tabLogin = getJsCookieLogin(contentDoc.defaultView);
  console.assert(tabLogin.isLoggedIn, "documentCookieSetter not logged in=" + tabLogin.toString()); // login cancelado, mas document já foi interceptado
  Cookies.setCookie(tabLogin, contentDoc.documentURIObject, obj.value, true);
}


function documentCookieGetter(obj, contentDoc) {
  //var tabLogin = TabLogin.getFromDomWindow(contentDoc.defaultView);
  //console.assert(tabLogin !== null, "not a tab document?");
  //if (tabLogin.isLoginInProgress) {
  //  tabLogin = TabLogin.getLoginInProgress(tabLogin.tabElement);
  //}

  var tabLogin = getJsCookieLogin(contentDoc.defaultView);
  console.assert(tabLogin.isLoggedIn, "documentCookieGetter not logged=" + tabLogin.toString() + "," + contentDoc.documentURI);

  var cookie2 = Cookies.getCookie(true, contentDoc.documentURIObject, tabLogin);
  var cookie = cookie2 === null ? "" : cookie2;
  return cookie; // send cookie value to content
}


/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Multifox.
 *
 * The Initial Developer of the Original Code is
 * Jeferson Hultmann <hultmann@gmail.com>
 * Portions created by the Initial Developer are Copyright (C) 2009
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */


const util2 = {

  logEx: function() {
    var buf = "logEx";// + this.logEx.caller + "\n";// + util2.throwStack._toString(false, Components.stack) + "===";
    for (var idx = 0, len = arguments.length; idx < len; idx++) {
      buf += "\n" + arguments[idx];// + "=" + this.caller[arguments[idx]];
    }

    buf += "\n====\n" + util2.throwStack._toString(false, Components.stack);
    console.log(buf);
  },

  throwStack: {

    go: function(txt) {
      console.log("Stack: " + txt + "\n" + this._toString(false, Components.stack));
      throw new Error(txt);
    },

    _toArray: function(webOnly, theStack) {
      var allItems = [];
      if (!theStack) {
        return allItems;
      }

      function StackItem() {}
      StackItem.prototype = {
        lang: "?",
        filename: "?",
        lineNumber: -1,
        name: "?"
      };

      // aparentemente é um bug que ocorre as vezes, todas as propriedades sao undefined
      if (theStack.languageName == undefined) {
        var item = new StackItem();
        item.name = theStack.toString();
        allItems.push(item);
      }

      // myStack.caller é quase sempre null, mas algumas vezes é undefined...
      for (var myStack = theStack; myStack; myStack = myStack.caller) {
        if (webOnly) {
          var n = myStack.filename;
          if (n === null) {
            continue;
          }

          if ((n.indexOf("http:") !== 0) && (n.indexOf("https:") !== 0)) {
            continue;
          }
        }


        var item = new StackItem();
        allItems.push(item);
        item.lang = myStack.languageName;
        item.filename = myStack.filename;
        item.lineNumber = myStack.lineNumber; // Valid line numbers begin at "1". "0" indicates unknown.
        item.name = myStack.name;
      }

      allItems.reverse();
      return allItems;
    },

    _toString: function(webOnly, theStack) {
      var arr = this._toArray(webOnly, theStack);
      var lines = new Array(arr.length);
      for (var idx = arr.length - 1, idx2 = 0; idx > -1; idx--) {
        var lang = webOnly ? "" : arr[idx].lang + " ";
        lines[idx2] = "[" + (idx + 1) + "] "
                    + lang
                    + arr[idx].filename + " "
                    + arr[idx].name + " ("
                    + arr[idx].lineNumber + ")";
        idx2++;
      }
      return lines.join("\n");
    }

  }
};




function endsWith(sufix, str) { // TODO usar hasRootDomain!!!
  var idx = str.lastIndexOf(sufix);
  if (idx === -1) {
    return false;
  }
  var idxMatch = str.length - sufix.length;
  return idx === idxMatch;
}



const InternalHost = {

  // .youtube.com ==> .youtube.com.<hexstrip>.multifox-anon-ns

  // .youtube.com ==> .youtube.com.<hex(youtube.com)>-<hex(user)>-<hex(tld)>.multifox-login-ns
  // www.yimg.com ==> www.yimg.com.<hex(www.yimg.com)>-<hex(user)>-<hex(tld)>.multifox-anon-ns
  //                                     ^^^-- TODO usar tld?


/*
  decode: function(hex) {
    return hexToString(hex);
  },

  encode: function(txt) {
    return stringToHex(txt);
  },
*/


  isInternalLoggedIn: function(host) { // used by onCookieChanged
    return hasRootDomain("multifox-auth-1", host);
  },

  getLoggedCookies: function() {
    var all = getAllCookiesFromHost("multifox-auth-1");
    return all;
  },

  getUserCookies: function(encUser, encTld) {
    var rv = [];
    var hexlogin = "-" + encUser + "-" + encTld;
    this._getUserCookies(rv, hexlogin, "multifox-auth-1");
    this._getUserCookies(rv, hexlogin, "multifox-anon-1");
    console.log("getUserCookies", rv.length);
    return rv;
  },

  _getUserCookies: function(rv, hexlogin, ns) {
    var all = getAllCookiesFromHost(ns);
    var cookie;
    var suffix = hexlogin + "." + ns;
    for (var idx = all.length - 1; idx > -1; idx--) {
      cookie = all[idx];
      if (endsWith(suffix, cookie.host)) {
        rv.push(cookie);
      }
    }
  },


  parseHost_enc: function(internalHost) {
    var a = internalHost.split(".").reverse();
    console.assert(a[0] === "multifox-auth-1", "invalid 87");
    var strip = a[1].split("-");
    if (strip.length !== 3) {
      return null;
    }
    return {
      enc_data: a[1],
      enc_realHostTld:  strip[0],
      enc_myUser: strip[1],
      enc_myTld:  strip[2]
    };
  },


  //getRealHost: function(internalHost) {
  parseHost: function(internalHost) { // TODO users de parseHost soh querem realHost
    var suf = ".multifox-auth-1";
    if (endsWith(suf, internalHost) === false) {
      console.assert(endsWith(".multifox-anon-1", internalHost) === false, "invalid 88");
      return null;
    }

    var a = internalHost.split(".").reverse();
    console.assert(a[0] === "multifox-auth-1", "invalid 87");
    var strip = a[1].split("-");
    console.assert(strip.length === 3, "invalid 89");
    var host = a.slice(2).reverse().join(".");

    //console.log('InternalHost.parseHost', internalHost, strip, host);

    return {
      // TODO retornar tb realHost TLD? host = getTldFromHost(hostData.realHost);
      realHost:  host,
      myUser: hexToString(strip[1]),
      myTld:  hexToString(strip[2])
    };
  }

};




const TabLogin = { //TabLoginStatic IdProvider

  NewAccount: "",

  create: function(tabElement, encUser, encTld) {
    return new TabInfo(tabElement, encUser, encTld);
  },

  getLoginInProgress: function(tab) { // BUG should new TabLogin be aware it is LoginInProgress?
    //var tab = this.tabElement;
    if (tab.hasAttribute("multifox-login-submit-domain")) { // BUG tab undef ao verificar update
      var usr = tab.getAttribute("multifox-login-submit-user")
      var tld = tab.getAttribute("multifox-login-submit-domain");
      return TabLogin.create(tab, stringToHex(usr), stringToHex(tld));
    } else {
      return null;//TabLogin.create(tab, null, null);
    }
  },

  removeLoginInProgress: function(tab) {
    var moveMode = tab.getAttribute("multifox-login-submit-data-moved");
    tab.removeAttribute("multifox-login-submit-domain");
    tab.removeAttribute("multifox-login-submit-user");
    tab.removeAttribute("multifox-login-submit-data-moved");
    return moveMode;
  },

  setLoginInProgress: function(username, host, tab) {
    console.assert(username.length > 0, "no username");
    tab.setAttribute("multifox-login-submit-user", username);
    tab.setAttribute("multifox-login-submit-domain", getTldFromHost(host));
  },

  getFromDomWindow: function(contentWin) {
    if (contentWin === null) {
      return null;
    }

    var url = contentWin.document.documentURI;
    if (url.length === 0) {
      return null;
    }

    var tab = WindowParents.getTabElement(contentWin);
    return (tab === null) ? null : new TabInfo(tab);
  }

};



function TabInfo(tabElement, usr, tld) {
  console.assert(tabElement !== null, "tab is invalid");

  if (usr === undefined) {
    console.assert(tld === undefined, "usr=undefined");
    if (tabElement.hasAttribute("multifox-tab-id-provider-tld-enc")) {
      usr = tabElement.getAttribute("multifox-tab-id-provider-user-enc");
      tld = tabElement.getAttribute("multifox-tab-id-provider-tld-enc");
    } else {
      usr = null;
      tld = null;
    }
  }
  this._userEncoded = usr;
  this._tldEncoded = tld;
  this.tabElement = tabElement;
}


TabInfo.prototype = {

  formatUri: function(uri) {
    var u = uri.clone();
    u.host = this.formatHost(u.host);
    return u;
  },

  formatHost: function(host) {
    if (this.isNewUser) {
      //console.log('toInternalHost TabLogin.NewAccount', host, this._tldEncoded, this._userEncoded);
      return host;
    }

    console.assert(typeof host === "string", "invalid domain2 ="+host);
    console.assert(typeof this._userEncoded === "string", "invalid user=" + this._userEncoded);
    console.assert(typeof this._tldEncoded === "string", "invalid loginTld=" + this._tldEncoded);
    //console.assert(this._userEncoded.indexOf(".${INTERNAL__DOMAIN__SUFFIX}") === -1, "invalid user");

    // We need to use tld(host) ==> otherwise, we couldn't (easily) locate the cookie for different subdomains
    // TODO perf mandar TLD com params
    var hex = stringToHex(getTldFromHost(host)) + "-" + this._userEncoded + "-" + this._tldEncoded;
    var suffix = this.isExternalAnonResource ? ".multifox-anon-1" : ".multifox-auth-1";
    return host + "." + hex + suffix;
  },


  equals: function(tabLogin) {
    return (tabLogin._userEncoded === this._userEncoded) && (tabLogin._tldEncoded === this._tldEncoded);
  },

  toString: function() {
    return "loginTld=" + this.plainTld + " loginUser=" + this.plainUser;
  },

  get plainTld() {
    return this._tldEncoded === null ? null : hexToString(this._tldEncoded);
  },

  get plainUser() {
    return this._tldEncoded === null ? null : hexToString(this._userEncoded);
  },

  get encodedTld() {
    return this._tldEncoded;
  },

  get encodedUser() {
    return this._userEncoded;
  },


  get hasUser() {
    return this._userEncoded !== null;
  },

  get isNewUser() {
    return this._userEncoded === TabLogin.NewAccount;
  },

  get isLoggedIn() {
    var usr = this._userEncoded;
    return (usr !== null) && (usr !== TabLogin.NewAccount);
  },

  getTabTld: function() {
    // use "location", "documentURI" may not reflect changes in hash (e.g. Twitter)
    return getTldFromHost(this.tabElement.linkedBrowser.contentDocument.location.hostname);
  },
  getTabTldEncoded: function() {
    return stringToHex(this.getTabTld());
  },
/*
  get loginObject
  updateIcon:
  setKey -- attr
*/
  toAnon: function() { // toUnlogged
    var tabLogin = new TabInfo(this.tabElement, this.encodedUser, this.encodedTld);
    tabLogin.anonResource = true;
    Object.freeze(tabLogin);
    return tabLogin;
  },

  get isExternalAnonResource() {
    return "anonResource" in this;
  },


  setTabAsAnon: function() {
    var tab = this.tabElement;
    console.trace("setTabAsAnon " + tab.getAttribute("multifox-tab-id-provider-tld-enc"));
    if (tab.hasAttribute("multifox-tab-id-provider-tld-enc")) {
      tab.removeAttribute("multifox-tab-id-provider-tld-enc");
      tab.removeAttribute("multifox-tab-id-provider-user-enc");
      invalidateUI(tab);
    }
  },

  saveToTab: function() { //apply
    var tab = this.tabElement;
    if (tab.hasAttribute("multifox-tab-error")) {
      // reset error icon
      tab.removeAttribute("multifox-tab-error");
    }

    var domain = this._tldEncoded;
    var user = this._userEncoded;

    //console.trace("commit login " + user + " at " + domain);

    console.assert(domain !== null, "setLogin tld=null");
    console.assert(typeof(domain) === "string", "setLogin tld="+domain);
    console.assert(typeof(user) === "string", "setLogin user="+user);

    if (tab.hasAttribute("multifox-tab-id-provider-tld-enc") === false) {
      tab.setAttribute("multifox-logging-in", "true"); // TODO
    }

    tab.setAttribute("multifox-tab-id-provider-tld-enc", domain);
    tab.setAttribute("multifox-tab-id-provider-user-enc", user);

    invalidateUI(tab);
  },


  get isLoginInProgress() {
    return this.tabElement.hasAttribute("multifox-login-submit-domain");
  },

  get hasLoginInProgressMoveData() {
    return this.tabElement.hasAttribute("multifox-login-submit-data-moved");
  },

  setLoginInProgressMoveData: function(mode) {
    this.tabElement.setAttribute("multifox-login-submit-data-moved", mode);
  }

};


const WindowParents = {

  getTabElement: function(contentWin) {
    var chromeWin = this.getChromeWindow(contentWin);
    if ((chromeWin !== null) && ("getBrowser" in chromeWin)) {
      var elem = chromeWin.getBrowser();
      switch (elem.tagName) {
        case "tabbrowser":
          var topDoc = contentWin.top.document;
          var idx = elem.getBrowserIndexForDocument(topDoc);
          if (idx > -1) {
            return getTabs(elem)[idx];
          }
          break;
        default: // view-source => tagName="browser"
          // Note that this file uses document.documentURI to get
          // the URL (with the format from above). This is because
          // document.location.href gets the current URI off the docshell,
          // which is the URL displayed in the location bar, i.e.
          // the URI that the user attempted to load.
          console.log("getTabElement=" + elem.tagName + "\n" +
                       contentWin.document.documentURI+ " " +chromeWin.document.location +"\n"+
                       contentWin.document.location   + " " +chromeWin.document.documentURI);
          break;
      }
    }
    return null;
  },

  getChromeWindow: function(contentWin) { // getTopWindowElement
    if (!contentWin) console.trace('contentWin='+contentWin);
    var qi = contentWin.QueryInterface;
    if (!qi) {
      return null;
    }

    if (contentWin instanceof Ci.nsIDOMChromeWindow) {
      // extensions.xul, updates.xul ...
      return contentWin;
    }

    var win = qi(Ci.nsIInterfaceRequestor)
                .getInterface(Ci.nsIWebNavigation)
                .QueryInterface(Ci.nsIDocShell)
                .chromeEventHandler
                .ownerDocument
                .defaultView;
    // wrappedJSObject allows access to gBrowser etc
    // wrappedJSObject=undefined sometimes. e.g. contentWin=about:multifox
    var unwrapped = win.wrappedJSObject; // TODO XPCNativeWrapper.unwrap?
    return unwrapped ? unwrapped : win;
  }

};




const OpenerWindowIdentity = { // TODO move to TabLogin?

  findFromWindow: function(contentWin) {
    // popup via js/window.open?
    if (contentWin.opener) {
      //console.log("_getFromOpenerContent opener=" + contentWin.opener.document.location);
      var tabOpener = WindowParents.getTabElement(contentWin.opener);
      if (tabOpener) {
        return new TabInfo(tabOpener);
      }
    }

    //console.log("_getFromOpenerContent - no opener=" + contentWin.opener);

    // fx starting ==> opener=null
    return this.findFromOpenerSelectedTab(contentWin); // id from selected tab
  },


  findFromOpenerSelectedTab: function(contentWin) {
    var chromeWin = WindowParents.getChromeWindow(contentWin);
    if (chromeWin && chromeWin.opener) {
      var type = chromeWin.opener.document.documentElement.getAttribute("windowtype");
      if (type === "navigator:browser") {
        var selTab = chromeWin.opener.getBrowser().selectedTab;
        return new TabInfo(selTab);
      }
    }

    return null;
  }
};



function getTabs(tabbrowser) {
  return tabbrowser.mTabContainer.childNodes;
}


// var serialize = {
function getUtf8Converter() {
  var conv = Cc["@mozilla.org/intl/scriptableunicodeconverter"].createInstance(Ci.nsIScriptableUnicodeConverter);
  conv.charset = "UTF-8";
  return conv;
}


function stringToBytes(str) {
  return getUtf8Converter().convertToByteArray(str, {});
}


function bytesToString(bytes) {
  return getUtf8Converter().convertFromByteArray(bytes, bytes.length);
}



function stringToHex(str) {
  var bin = stringToBytes(str); // "€".charCodeAt(0) = 8364
  var len = bin.length;
  var hex = new Array(len * 2);

  var j = 0;
  var alphabet = "abcdefghijklmnop";

  for (var idx = 0; idx < len; idx++) {
    var myByte = bin[idx];
    var nibble1 = (myByte & 0xf0) >> 4;
    var nibble0 =  myByte & 0x0f;
    hex[j++] = alphabet[nibble1];
    hex[j++] = alphabet[nibble0];
  }

  return hex.join("");
}


function hexToString(hex) { // stringDecode
  console.assert(typeof hex === "string", "hexToString val=" + hex);
  var len = hex.length / 2;
  if (len === 0) {
    return "";
  }

  console.assert((len % 1) === 0, "invalid hex string: " + hex);
  var bin = new Array(len);

  var j = 0;
  var offset = 97; // 97="abcdefghijklmnop".charCodeAt(0)

  for (var idx = 0; idx < len; idx++) {
    var nibble1 = hex.charCodeAt(j++) - offset;
    var nibble0 = hex.charCodeAt(j++) - offset;
    bin[idx] = (nibble1 << 4) | nibble0;
  }

  return bytesToString(bin);
}



function getTldFromHost(host) {
  /*
 if (host.indexOf(":") > - 1) {
    return host; // literal ipv6 (e.g. 3ffe:2a00:100:7031::1)
  }
  var first = host.charCodeAt(0);
  if ((first >= 48) && (first <= 57)) { // "0123456789"
    return host; // literal ipv4 (e.g. 127.0.0.1, 0x7f.0.0.1)
  }
  */

  var firstDot = host.indexOf(".");
  if (firstDot === -1) {
    return host; // "localhost"
  }

  // firstDot=0 ("...local.host") (e.g. from cookies)
  // OBS "..local.host" return "localhost"
  if (firstDot === 0) {
    return getTldFromHost(host.substr(1)); // recursive
  }

  //  "localhost" ==> -1 exception
  // ".localhost" ==>  0 exception
  // ".loc.al.ho.st" ==> 0 exception
  //  "local.host" ==> >0 OK
  try {
    return Services.eTLD.getBaseDomainFromHost(host);
  } catch (ex) {
    if (ex.result !== Components.results.NS_ERROR_HOST_IS_IP_ADDRESS) {
      console.trace(host + "/" + ex);
    }
    return host; // literal ipv4/ipv6?
  }
}



/*

Cu.import("resource://gre/modules/Services.jsm");
Services.eTLD.getBaseDomainFromHost("10.11.12.13");

const Cc = Components.classes;
const Ci = Components.interfaces;

function getUtf8Converter() {
  var conv = Cc["@mozilla.org/intl/scriptableunicodeconverter"].createInstance(Ci.nsIScriptableUnicodeConverter);
  conv.charset = "UTF-8";
  return conv;
}

function bytesToString(bytes) {
  return getUtf8Converter().convertFromByteArray(bytes, bytes.length);
}


function hexToString(hex) {
  var len = hex.length / 2;
  var bin = new Array(len);

  var j = 0;
  var offset = 97;

  for (var idx = 0; idx < len; idx++) {
    var nibble1 = hex.charCodeAt(j++) - offset;
    var nibble0 = hex.charCodeAt(j++) - offset;
    bin[idx] = (nibble1 << 4) | nibble0;
  }

  return bytesToString(bin);
}

hexToString('gmgjgoglgfgegjgocogdgpgn');

multifox[03:57:17.781] LoginData.setDefaultLogin user not found   ghgpgpghgmgfcogdgpgn ["gihfgmhegngbgogo"]

gggbgdgfgcgpgpglcogdgpgn gihfgmhegngbgogoeaghgngbgjgmcogdgpgn gggbgdgfgcgpgpglcogdgpgn               hultmann@gmail.com facebook.com
gggbgdgfgcgpgpglcogdgpgn gkgigmhegngogodcdc                   hehhgjhehegfhccogdgpgn    facebook.com jhltmnn22 twitter.com
gggbgdgfgcgpgpglcogdgpgn gchcfpgngphkgegfhg                   hehhgjhehegfhccogdgpgn                 br_mozdev twitter.com


gggbgdgfgcgpgpglcogdgpgn-gihfgmhegngbgogo                    -hdhegbhegdgphfgohegfhccogdgpgn statcounter.com
gggbgdgfgcgpgpglcogdgpgn-gihfgmhegngbgogoeaghgngbgjgmcogdgpgn-gggbgdgfgcgpgpglcogdgpgn
*/